package iSATLibrary;

import iSATLibrary.ISatLibrary.IPASIR;
import iSATLibrary.ISatLibrary.ISATLIB;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;

import com.sun.jna.Callback;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.IntByReference;

class SolverLibrary {
	protected IPASIR ipasirLibrary;
	protected ISATLIB isatLibrary;

	SolverLibrary(IPASIR ipasirLibrary, ISATLIB isatLibrary) {
		this.ipasirLibrary = ipasirLibrary;
		this.isatLibrary = isatLibrary;
	}
}

/**
 * This class is for using incremental SAT library API (IPASIR)
 * and its extended API available from Java. This wraps these APIs.
 * Use the methods of this class when using incremental SAT APIs.
 * 
 * @author Tatsuya Sako
 */
public class SatSolverForNative extends ISatLibrary implements ISatSolver{

	public interface ICallback extends Callback {
		public int invoke(IntByReference state);
	}

	public class UserCallback implements ICallback {
		public int invoke(IntByReference state) {
			return state.getPointer().getInt(0) != 0 ? 1 : 0;
		}
	}

	private ArrayList<int[]> buffer = new ArrayList<int[]>(); // temporary buffer for the clauses to add to SAT solver
	private int countOfClauses = 0;
	private int bufsize; // upper limit number of clauses to add SAT solver at once
	private String solverName;
	private IPASIR ipasirLib;
	private ISATLIB isatLib;
	private Pointer solverPtr;

	// For callback
	private UserCallback callback;
	private IntByReference state;
	private Pointer statePtr;

	// constructor
	public SatSolverForNative(String name) {
		super(name);
		this.solverName = super.solverName;
		this.ipasirLib = super.ipasirLib;
		this.isatLib = super.isatLib;
		this.solverPtr = super.solverPtr;
		this.bufsize = 1000;
		callback = new UserCallback();
		state = new IntByReference();
		statePtr = state.getPointer();
		this.setTerminate(0);
	}

	/**
	 * Gets a signature of SAT solver.
	 * 
	 * @return
	 * the signature of SAT solver
	 */
	public String getSignature() {
		return ipasirLib.ipasir_signature();
	}

	/**
	 * Constructs a new SAT solver.
	 * 
	 * @return
	 * 	solver's object
	 */
	public SatSolverForNative init() {
		SatSolverForNative solver = new SatSolverForNative(solverName);
		return solver;
	}

	/**
	 * Releases the solver.
	 */
	public void release() {
		ipasirLib.ipasir_release(solverPtr);
	}

	/**
	 * Adds a literal to SAT solver or finalizes the clause with a "0".
	 * 
	 * @param lit_or_zero
	 *  a literal or "0"
	 */
	public void add(int lit_or_zero) {		
		ipasirLib.ipasir_add(solverPtr, lit_or_zero);
	}

	/**
	 * Adds an assumption.
	 * 
	 * @param lit
	 *  the assumption
	 */
	public void assume(int lit) {		
		ipasirLib.ipasir_assume(solverPtr, lit);
	}

	/**
	 * Solves a SAT problem.
	 * 
	 * @return
	 * true (if the problem is SAT), false (UNSAT).
	 */
	public int solve() {
		done();
		return ipasirLib.ipasir_solve(solverPtr);
	}

	/**
	 * Gets the truth value of the given literal in the assignment.
	 * 
	 * @param lit
	 *  the literal
	 * @return
	 * the truth value of the given literal
	 */
	public int val(int lit) {
		return ipasirLib.ipasir_val(solverPtr, lit);
	}

	/**
	 * Checks if the given assumption was used to prove UNSAT in the last search.
	 * 
	 * @param lit
	 *  the assumption literal
	 * @return
	 * 	1 (used), 0 (otherwise)
	 */
	public int failed(int lit) {
		return ipasirLib.ipasir_failed(solverPtr, lit);
	}

	/**
	 * Sets a callback function requiring a termination of solving to the solver.
	 * 
	 * @param value
	 */
	public void setTerminate(int value) {
		if (value == 0)
			ipasirLib.ipasir_set_terminate(solverPtr, null, null);
		else {
			state.setValue(value);
			ipasirLib.ipasir_set_terminate(solverPtr, statePtr, callback);
		}
	}

	/**
	 * Sets a number of clauses added to SAT solver at a time. 
	 * Default value is 1000.
	 * 
	 * @param size
	 *  the number of clauses added to SAT solver at a time
	 */
	public void setBufferSize(int size) {
		this.bufsize = size;
	}

	/**
	 * Makes an array for buffering clauses.
	 * This method appends a new clause to the end of the buffer.
	 * 
	 * @return
	 * appended buffer
	 */
	public int[] makeBuffer() {
		int length = 0;
		int offset = 0;
		for (int[] array : buffer) {
			length += (array.length + 1); // for adding "0"
		}
		int[] margedBuffer = new int[length];
		for (int[] array : buffer) {
			int l = array.length;
			for (int i = 0; i < l; i++) {
				margedBuffer[offset + i] = array[i];
			}
			margedBuffer[offset + l] = 0;
			offset = offset + l + 1;
		}
		buffer.clear();
		return margedBuffer;
	}

	/**
	 * Sets the verbosity of SAT solver's output.
	 * If you set "0", solver's output isn't displayed.
	 * 
	 * @param verbosity
	 *  the verbosity of SAT solver's output
	 */
	public void setVerbosity(int verbosity) {
		isatLib.isat_set_verbosity(solverPtr, verbosity);
	}

	@Override
	public void setVars(int count) {
	}
	
	

	/**
	 * Gets the number of variables in SAT solver.
	 * 
	 * @return
	 *  number of variables
	 */
	public int getVars() {
		return isatLib.isat_vars(solverPtr);
	}

	/**
	 * Gets the number of clauses in SAT solver.
	 * 
	 * @return
	 *  number of clauses
	 */
	public int getClauses() {
		return isatLib.isat_clauses(solverPtr);
	}
	
	public void newVar(int howmany){
		isatLib.isat_new_var(solverPtr, howmany);
	}

	/**
	 * Adds a clause to SAT solver.
	 * You must add "0" to the end of the array "clause".
	 * 
	 * @param clause
	 *  a clause to add to SAT solver
	 */
	public void add_a_clause(int[] clause) {
		isatLib.isat_add_clauses(solverPtr, clause, clause.length);
	}

	/**
	 * Adds multiple clauses to SAT solver at a time.
	 * Clauses are temporally buffered.
	 * 
	 * @param lits
	 *  array of literals
	 */
	public void addNormalizedClause(int[] lits) {
		buffer.add(lits.clone());
		countOfClauses++;
		if (countOfClauses == bufsize) {
			int[] clauses = makeBuffer();
			add_a_clause(clauses);
			countOfClauses = 0;
		}
	}

	/**
	 * Adds a clause to SAT solver.
	 * 
	 * @param clause
	 *  clause to add to SAT solver
	 */
	public void addClause(int[] clause) {
		addNormalizedClause(clause);
	}
	
	/**
	 * Adds clauses to SAT solver.
	 * 
	 * @param clauses
	 *  clauses to add to SAT solver
	 */
	public void addClauses(int[][] clauses) {
		for(int[] clause : clauses)
			addNormalizedClause(clause);
	}
	
	/**
	 * Adds assumption literals to SAT solver.
	 * 
	 * @param literals
	 *  literals to be assumption 
	 */
	public void addAssumptions(int[] literals) {
		isatLib.isat_add_assumptions(solverPtr, literals, literals.length);
	}

	/**
	 * Adds rest clauses in buffer to SAT solver.
	 * This method is called before solving a SAT problem.
	 */
	public void done() {
		if (countOfClauses == 0)
			return;
		int[] clauses = makeBuffer();
		add_a_clause(clauses);
		countOfClauses = 0;
	}

	/**
	 * Solves a SAT problem with assumption.
	 * 
	 * @param assumptions
	 *  assumption
	 * @return
	 *  true (if the problem is satisfiable(SAT) ), false (otherwise).
	 */
	public int solve(int[] assumptions) {
		if (assumptions != null) {
			for (int lit : assumptions)
				ipasirLib.ipasir_assume(solverPtr, lit);
		}
		return solve();
	}

	/**
	 * Adds a blocking clause to SAT solver.
	 * 
	 * @param clause
	 *  the clause required to block
	 */
	public void addBlockingClause(int[] clause) {
		isatLib.isat_add_blocking_clause(solverPtr, clause, clause.length);
	}

	/**
	 * Adds a blocking clause to SAT solver.
	 * This blocking clause contains all SAT variables and blocks obtained solution only.
	 */
	public void addBlockingClause() {
		isatLib.isat_add_blocking_clause(solverPtr, null, 0);
	}

	/**
	 * Adds a pseudo boolean constraint to SAT solver.
	 * PB constraint is encoded CNF and clauses generated from the constraint are added to SAT solver inside.
	 * @param weights 
	 *  each coefficient of literals
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of PB constraint 
	 * @param isgeq
	 *  true if added constraint is  [... >= k], false otherwise
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addPB(int[] weights, int[] literals, int k, boolean isgeq){
		if(weights.length != literals.length){
			System.out.println("Error : weights and literals are must be same size.");
			return -1;
		}
		return isatLib.isat_add_PB(solverPtr, weights, literals, literals.length, k, isgeq);
	}
	
	/**
	 * Adds a pseudo boolean constraint to SAT solver.
	 * PB constraint is encoded CNF and clauses generated from the constraint are added to SAT solver inside.
	 * @param weights 
	 *  each coefficient of literals
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of PB constraint 
	 * @param isgeq
	 *  true if added constraint is [... >= k], false otherwise
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addPBWithAssumption(int[] weights, int[] literals, int k, boolean isgeq, int assumption){
		if(weights.length != literals.length){
			System.out.println("Error : weights and literals are must be same size.");
			return -1;
		}
		return isatLib.isat_add_PB_with_assumption(solverPtr, weights, literals, literals.length, k, isgeq, assumption);
	}
	
	/**
	 * Adds an at most k constraint to SAT solver.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addAtMost(int[] literals, int k){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		return isatLib.isat_add_PB(solverPtr, weights, literals, length, k, false);
	}

	/**
	 * Adds an at most k constraint to SAT solver as a relaxed constraint.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addAtMostWithAssumption(int[] literals, int k, int assumption){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		return isatLib.isat_add_PB_with_assumption(solverPtr, weights, literals, length, k, false, assumption);
	}

	/**
	 * Adds an at least k constraint to SAT solver.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addAtLeast(int[] literals, int k){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		return isatLib.isat_add_PB(solverPtr, weights, literals, length, k, true);
	}
	
	/**
	 * Adds an at least k constraint to SAT solver as a relaxed constraint.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addAtLeastWithAssumption(int[] literals, int k, int assumption){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		return isatLib.isat_add_PB_with_assumption(solverPtr, weights, literals, length, k, true, assumption);
	}

	/**
	 * Adds an exact k constraint to SAT solver.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addExactly(int[] literals, int k){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		isatLib.isat_add_PB(solverPtr, weights, literals, length, k, false);
		return isatLib.isat_add_PB(solverPtr, weights, literals, length, k, true);
	}
	
	/**
	 * Adds an exact k constraint to SAT solver as a relaxed constraint.
	 * @param literals
	 *  literals
	 * @param k
	 *  integer of right side of at least constraint
	 * @return
	 *  biggest auxiliary variable
	 */
	public int addExactlyWithAssumption(int[] literals, int k, int assumption){
		int length = literals.length;
		int[] weights = new int[length];
		Arrays.fill(weights, 1);
		isatLib.isat_add_PB_with_assumption(solverPtr, weights, literals, length, k, false, assumption);
		return isatLib.isat_add_PB_with_assumption(solverPtr, weights, literals, length, k, true, assumption);
	}

	/**
	 * Excludes a literal from the simplification of SAT solver.
	 * This method is available only when SAT solver uses simplification process.
	 * If you use the SAT solver which does "simplification", you should freeze the literal.
	 * 
	 * @param lit
	 *  the literal required to freeze
	 */
	public void setFrozen(int lit) {
		isatLib.isat_freeze(solverPtr, lit);
	}

	/**
	 * Checks if the literal is frozen.
	 * This method is available only when SAT solver uses simplification process.
	 * 
	 * @param lit
	 *  the literal required to check if frozen
	 * @return
	 *  true (if the literal is frozen), false (otherwise)
	 */
	public boolean isEliminated(int lit) {
		return isatLib.isat_is_eliminated(solverPtr, lit);
	}

	/**
	 * Gets SAT solution (assignments to SAT variables) by bitset.
	 * If you need to decode to CSP solution by Sugar, you should use this method.
	 * 
	 * @return
	 *  the model of solution by bitset
	 */
	public BitSet getModel() {
		BitSet model = new BitSet();
		int value;
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for (int i = 0, v = isatLib.isat_vars(solverPtr); i < v; i++) {
			value = p.getInt(i * 4);
			if (value > 0)
				model.set(value);
		}
		isatLib.isat_clean(solverPtr, p);
		return model;
	}

	/**
	 * Gets SAT solution (assignments to SAT variables) by binary integer array.
	 * 
	 * @return
	 *  the model of solution by binary integer array (true:1, false:0)
	 */
	public int[] get_int_model_by_binary() {
		int length = (int) isatLib.isat_vars(solverPtr);
		int[] model = new int[length];
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for (int i = 0; i < length; i++)
			model[i] = p.getInt(i * 4) > 0 ? 1 : 0;
		isatLib.isat_clean(solverPtr, p);
		return model;
	}

	/**
	 * Gets SAT solution (assignments to SAT variables) by integer array.
	 * 
	 * @return
	 *  the model of solution by integer array
	 */
	public int[] getIntModel() {
		int length = (int) isatLib.isat_vars(solverPtr);
		int[] model = new int[length];
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for (int i = 0; i < length; i++)
			model[i] = p.getInt(i * 4);
		isatLib.isat_clean(solverPtr, p);
		return model;
	}

	/**
	 * Get results of assumption literals if used proving unsatisfiability.
	 * 
	 * @param literals
	 *  assumption literals
	 * @return
	 *  the result of assumption literals(1:used, 0:not used)
	 */
	public int[] getFailed(int[] literals) {
		int length = literals.length;
		int[] result = new int[length];
		IntByReference resptr = new IntByReference();
		resptr = isatLib.isat_failed(solverPtr, literals, length);
		Pointer p = resptr.getPointer();
		for (int i = 0; i < length; i++)
			result[i] = p.getInt(i * 4);
		isatLib.isat_clean(solverPtr, p);
		return result;
	}
	
	
	public int[] getMUS(){
		System.out.println("c This method is unavailable.");
		System.exit(1);
		return null;
	}
	
	/**
	 * Outputs SAT solution on standard output.
	 */
	public void outputModel() {
		BitSet model = getModel();
		for (int i = 1; i <= isatLib.isat_vars(solverPtr); i++)
			System.out.print(model.get(i) ? i + " " : -i + " ");
		System.out.println();
	}

	/**
	 * Gets total time of invocation of SAT solver.
	 * 
	 * @return
	 *  elapsed time
	 */
	public double getTime() {
		return isatLib.isat_get_time(solverPtr);
	}

	/**
	 * Gets a name of SAT solver.
	 * 
	 * @return
	 *  the name of SAT solver
	 */
	public String getSolverName() {		
		return solverName;
	}

}