package iSATLibrary;

import org.sat4j.core.Vec;
import org.sat4j.core.VecInt;
import org.sat4j.minisat.SolverFactory;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.TimeoutException;
import org.sat4j.tools.xplain.HighLevelXplain;
import org.sat4j.pb.tools.PBAdapter;

import static iSugar.Solver.Status.*;
import iSugar.Solver;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.BitSet;

import jp.kobe_u.sugar.Logger;

/**
 * This class is for solving problem by using sat4j as backend SAT solver.
 * The non-incremental search in enumeration and solving HCP is unsupported.
 * @author sako
 *
 */
public class SatSolverForSat4j implements ISatSolver {

	private String solverName = "sat4j";
	private HighLevelXplain<ISolver> solver;
	private PBAdapter pb = new PBAdapter(solver);
	private VecInt clauseToAdd = new VecInt();
	private VecInt assumptions = new VecInt();
	private ArrayList<ArrayList<Integer>> clauses = new ArrayList<ArrayList<Integer>>();
	private double startTime;
	public static int cIndex = 0;
	public boolean isusc = false;
	public static boolean isParamClause = false;
	public static int buffer = 0; // buffer for variables added to solver later
	
	// constructor
	public SatSolverForSat4j() {
		solver = new HighLevelXplain<ISolver>(SolverFactory.newDefault());
		isusc = Solver.strategy.equals("us1");
		startTime = System.nanoTime();
	}
		
	@Override
	public String getSignature() {
		/*
		URL url = AbstractLauncher.class.getResource("/sat4j.version"); //$NON-NLS-1$
		String signature = "sat4j";
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));
			signature = "sat4j " + in.readLine();
			in.close();
		} catch (IOException e) {
			System.out.println("c ERROR: " + e.getMessage());
		}*/
		String signature = "sat4j revision 2490";
		return signature;
	}

	@Override
	public ISatSolver init() {
		SatSolverForSat4j solver = new SatSolverForSat4j();
		return solver;
	}

	@Override
	public void release() {
		solver = null;
	}

	@Override
	public void add(int lit_or_zero) {
		if(lit_or_zero!=0)
			clauseToAdd.push(lit_or_zero);
		else {
			addClause(clauseToAdd.toArray());
			clauseToAdd.clear();
		}
	}

	@Override
	public void assume(int lit) {
		assumptions.push(lit);
	}

	@Override
	public int solve() {
		int res;
		//problem = solver;
		try {
			res = solver.isSatisfiable(assumptions) ? SATISFIABLE : UNSATISFIABLE;
			if(Logger.verboseLevel > 0)
				solver.printStat(new PrintWriter(System.out, true));
			assumptions.clear();
		} catch (TimeoutException e) {
			res = INDETERMINATE;
		}

		return res;
	}

	@Override
	public int val(int lit) {
		return solver.model(lit) ? lit : -lit;
	}

	@Override
	public int failed(int lit) {
		return solver.unsatExplanation().contains(lit) ? 1 : 0;
	}

	@Override
	public void setTerminate(int value) {
		if(value != 0)
			solver.expireTimeout();
	}

	@Override
	public String getSolverName() {
		return solverName;
	}

	@Override
	public void setVerbosity(int verbosity) {
		solver.setVerbose(verbosity > 0);
	}

	@Override
	public void setVars(int count) {
		solver.newVar(count+buffer);
	}

	@Override
	public int getVars() {
		return solver.nVars();
	}
	
	@Override
	public int getClauses() {
		return solver.nConstraints();
	}
	
	@Override
	public void newVar(int howmany){
		solver.newVar(howmany+buffer);
	}
	
	public void saveClause(int[] clause){
		ArrayList<Integer> c = new ArrayList<Integer>();
		for(int i : clause)
			c.add(i);
		clauses.add(c);
	}
	
	@Override
	public void addClause(int[] clause) {
		try{
			saveClause(clause);

			solver.addClause(new VecInt(clause), isusc&&!isParamClause ? cIndex : cIndex++);
		} catch (ContradictionException e){
			e.printStackTrace();
		}
	}
	
	@Override
	public void addClauses(int[][] clauses) {
		for(int[] clause : clauses)
			addClause(clause);
	}

	@Override
	public void addAssumptions(int[] lits) {
		if(lits == null) return;
		for(int lit : lits)
			assume(lit);
	}

	@Override
	public int solve(int[] assumptions) {
		done();
		addAssumptions(assumptions);
		return solve();
	}

	@Override
	public void addBlockingClause() {
		int[] model = solver.model();
		try {
			for(int i=0; i<model.length; i++)
				model[i] *= -1;
			//addClause(model);
			solver.addBlockingClause(new VecInt(model));
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void addBlockingClause(int[] clause) {
		try {
			for(int i=0; i<clause.length; i++)
				clause[i] *= -1;
			solver.addBlockingClause(new VecInt(clause));
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
	}

	@Override
	public int addPB(int[] weights, int[] literals, int k, boolean isgeq) {
		BigInteger d = BigInteger.valueOf(k);
		BigInteger[] coeffs = new BigInteger[weights.length];
		for(int i=0; i<weights.length; i++)
			coeffs[i] = BigInteger.valueOf(weights[i]);
		try {
			pb.addPseudoBoolean(new VecInt(literals), new Vec<BigInteger>(coeffs), isgeq, d);
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
		return solver.nVars();
	}

	@Override
	public int addPBWithAssumption(int[] weights, int[] literals, int k, boolean isgeq, int assumption) {
		System.out.println("method \"addPBWithAssumption\" is unimplemented for sat4j.");
		System.exit(1);
		return 0;
	}

	@Override
	public int addAtMost(int[] literals, int k) {
		try {
			solver.addAtMost(new VecInt(literals), k);
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
		return solver.nVars();
	}

	@Override
	public int addAtMostWithAssumption(int[] literals, int k, int assumption) {
		System.out.println("method \"addAtMostWithAssumption\" is unimplemented for sat4j.");
		System.exit(1);
		return 0;
	}

	@Override
	public int addAtLeast(int[] literals, int k) {
		try {
			solver.addAtLeast(new VecInt(literals), k);
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
		return solver.nVars();
	}

	@Override
	public int addAtLeastWithAssumption(int[] literals, int k, int assumption) {
		System.out.println("method \"addAtLeastWithAssumption\" is unimplemented for sat4j.");
		System.exit(1);
		return 0;
	}

	@Override
	public int addExactly(int[] literals, int k) {
		try {
			solver.addExactly(new VecInt(literals), k);
		} catch (ContradictionException e) {
			e.printStackTrace();
		}
		return solver.nVars();
	}

	@Override
	public int addExactlyWithAssumption(int[] literals, int k, int assumption) {
		System.out.println("method \"addExactlyWithAssumption\" is unimplemented for sat4j.");
		System.exit(1);
		return 0;
	}
	
	@Override
	public void setFrozen(int lit) {
		System.out.println("method \"setFrozen\" is unimplemented for sat4j.");
		System.exit(1);
	}

	@Override
	public boolean isEliminated(int lit) {
		System.out.println("method \"isEliminated\" is unimplemented for sat4j.");
		System.exit(1);
		return false;
	}

	@Override
	public BitSet getModel() {
		BitSet model = new BitSet();
		int[] m = solver.model();
		for(int val : m)
			if(val > 0)
				model.set(val);
		return model;
	}

	@Override
	public int[] getIntModel() {
		return solver.model();
	}

	@Override
	public int[] getFailed(int[] literals) {
		int length = literals.length;
		int[] result = new int[length];
		// return set of clause index which become UNSAT when assumption is assigned "true"
		int[] cause = solver.unsatExplanation().toArray();
		for(int i=0; i<length; i++){
			for(int j : cause){
				if (clauses.get(Math.abs(j)).contains(literals[i]) || clauses.get(Math.abs(j)).contains(-literals[i])){
					result[i] = 1;
					break;
				}
			}
		}
		return result;
	}
	
	public int[] getMUS(){
		int[] core = {};
		try {
			core = solver.minimalExplanation();
		} catch (TimeoutException e) {
			e.printStackTrace();
		}
		return core;
	}

	@Override
	public void outputModel() {
		BitSet model = getModel();
		for (int i = 1; i <= solver.nVars(); i++)
			System.out.print(model.get(i) ? i + " " : -i + " ");
		System.out.println();		
	}

	@Override
	public double getTime() {
		return (System.nanoTime()-startTime)/1000000000.0;
	}

	@Override
	public void addNormalizedClause(int[] clause) {
		addClause(clause);
	}
	
	@Override
	public void done(){
		if(Logger.verboseLevel > 0){
			System.out.println("c declared #vars     " + solver.nVars());
			if (solver.nVars() < solver.realNumberOfVariables()) {
				System.out.println("c internal #vars     " + solver.realNumberOfVariables());
			}	
			System.out.println("c #constraints  " + solver.nConstraints());
			solver.printInfos(new PrintWriter(System.out, true));
		}
	}
	
}
