package iSugar;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.csp.IntegerDomain;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.expression.Expression;

/**
 * This class is for encoding a channeling constraint ∑xi # v.
 * p_i are 0-1 variables and v is an integer variable.
 * @author sako
 *
 */
public class ChannelingConstraintEncoder {

	public static List<Expression> expressions;
	public static List<IntegerVariable> newVars;
	public static String encoding = "totalizer";
	
	private ChannelingConstraintEncoder(){
	}
	
	public static void encodeCC(List<IntegerVariable> vList, IntegerVariable v) throws SugarException, IOException{
		switch(encoding){
		case "seqcounter":
			encodeCCBySeqCounter(vList,v);
			break;
		case "totalizer":
			encodeCCByTotalizer(vList,v);
			break;
		default:
			throw new SugarException("encoding " + encoding + "doesn't exist.");
		}
	}
	
	/**
	 * translate an equation [v1 + v2 + ... +vn = v] to 2-terms equations by sequential counter method.
	 * v1 + v2 + v3 + v4 = v is translated below.
	 * s1 = v1;  s2 = s1 + v2;  s3 = s2 + v3;  s4 = s3 + v4;  v = s4
	 * vi is needed to be 0-1 variable.
	 * @param vList
	 *  the variables of left side 
	 * @param v
	 *  the variable of right side
	 * @throws SugarException
	 * 
	 * @throws IOException
	 * 
	 */
	public static List<IntegerVariable> encodeCCBySeqCounter(List<IntegerVariable> vList, IntegerVariable v) throws SugarException, IOException{
		IntegerVariable s1=null, s2, s3;
		expressions = new ArrayList<Expression>();
		List<Expression>exEq, exAdd;
		newVars = new ArrayList<IntegerVariable>();
		int offset = v.getDomain().getLowerBound();
		int ub = 0;
		while(!vList.isEmpty()){
			exEq = new ArrayList<Expression>();
			exAdd = new ArrayList<Expression>();
			s2 = vList.remove(0);
			if(s1 == null)
				ub = s2.getDomain().getUpperBound();
			else
				ub = s1.getDomain().getUpperBound() + s2.getDomain().getUpperBound();
			IntegerDomain domain = IntegerDomain.create(0, ub);
			s3 = new IntegerVariable(domain);
			newVars.add(s3);
			//exEq.add(Expression.LE); // EQ
			exEq.add(Expression.EQ); // EQ
			exEq.add(Expression.create(s3.getName()));
			if(s1 == null)
				exEq.add(Expression.create(s2.getName()));
			else {
				exAdd.add(Expression.ADD);
				exAdd.add(Expression.create(s1.getName()));
				exAdd.add(Expression.create(s2.getName()));
				exEq.add(Expression.create(new ArrayList<Expression>(exAdd)));
			}
			expressions.add(Expression.create(new ArrayList<Expression>(exEq)));
			s1 = s3;
		}
		exEq = new ArrayList<Expression>();
		exAdd = new ArrayList<Expression>();
		//exEq.add(Expression.LE); // EQ
		exEq.add(Expression.EQ); // EQ
		exEq.add(Expression.create(v.getName()));
		exAdd.add(Expression.ADD);
		exAdd.add(Expression.create(s1.getName()));
		exAdd.add(Expression.create(offset));
		exEq.add(Expression.create(new ArrayList<Expression>(exAdd)));
		expressions.add(Expression.create(new ArrayList<Expression>(exEq)));
		return newVars;
	}
	
	/**
	 * translate an equation [v1 + v2 + ... +vn = v] to 2-terms equations by totalizer method.
	 * v1 + v2 + v3 + v4 = v is translated below.
	 * s1 = v1 + v2;  s2 = v3 + v4;  s3 = s1 + s2;  v = s3
	 * vi is needed to be 0-1 variable.
	 * @param vList
	 *  the variables of left side 
	 * @param v
	 *  the variable of right side
	 * @throws SugarException
	 * 
	 * @throws IOException
	 * 
	 */
	public static List<IntegerVariable> encodeCCByTotalizer(List<IntegerVariable> vList, IntegerVariable v) throws SugarException, IOException{
		expressions = new ArrayList<Expression>();
		newVars = new ArrayList<IntegerVariable>();
		List<Expression> exEq = new ArrayList<Expression>();
		List<Expression> exAdd = new ArrayList<Expression>();
		int offset = v.getDomain().getLowerBound();
		IntegerVariable s = totalize(vList);
		//exEq.add(Expression.LE); // EQ
		exEq.add(Expression.EQ); // EQ
		exEq.add(Expression.create(v.getName()));
		exAdd.add(Expression.ADD);
		exAdd.add(Expression.create(s.getName()));
		exAdd.add(Expression.create(offset));
		exEq.add(Expression.create(new ArrayList<Expression>(exAdd)));
		expressions.add(Expression.create(new ArrayList<Expression>(exEq)));
		return newVars;
	}
	
	public static int calcVars(int dom){
		switch(encoding){
		case "totalizer":
			return calcVarsForTotalizer(dom);
		case "seqcounter":
			return calcVarsForSeqCounter(dom);
		default:
			return 0;
		}
	}
	
	public static int calcVarsForTotalizer(int dom){
		if(dom<1)
			return 0;
		else if(dom==1)
			return 1;
		else {
			int d1 = (dom+1)/2;
			int d2 = dom-d1;
			return dom + calcVarsForTotalizer(d1) + calcVarsForTotalizer(d2);
		}
	}
	
	public static int calcVarsForSeqCounter(int dom){
		return (dom*dom+3*dom)/2;
	}
	
	public static <T> List<List<T>> splitArray(List<T> list){
		 List<List<T>> lists = new  ArrayList<List<T>>();
		 List<T> l1 = new ArrayList<T>();
		 List<T> l2 = new ArrayList<T>();
		 int mid = (list.size()+1)/2;
		 for(int i=0; i<mid; i++)
			 l1.add(list.get(i));
		 for(int i=mid; i<list.size(); i++)
			 l2.add(list.get(i));
		 lists.add(l1); lists.add(l2);
		 return lists;
	}

	public static IntegerVariable totalize(List<IntegerVariable> vList) throws SugarException{
		List<Expression> exEq = new ArrayList<Expression>();
		switch(vList.size()){
		case 1:
			return vList.get(0);
		case 2:
			List<Expression> exAdd = new ArrayList<Expression>();
			IntegerVariable s1 = vList.get(0);
			IntegerVariable s2 = vList.get(1);
			int ub = s1.getDomain().getUpperBound() + s2.getDomain().getUpperBound();
			IntegerDomain domain = IntegerDomain.create(0, ub);
			IntegerVariable s3 = new IntegerVariable(domain);
			newVars.add(s3);
			//exEq.add(Expression.LE); // EQ
			exEq.add(Expression.EQ); // EQ
			exEq.add(Expression.create(s3.getName()));
			exAdd.add(Expression.ADD);
			exAdd.add(Expression.create(s1.getName()));
			exAdd.add(Expression.create(s2.getName()));
			exEq.add(Expression.create(new ArrayList<Expression>(exAdd)));
			expressions.add(Expression.create(new ArrayList<Expression>(exEq)));
			return s3;
		default:
			List<List<IntegerVariable>> lists = splitArray(vList);
			List<IntegerVariable> newList = new ArrayList<IntegerVariable>();
			newList.add(totalize(lists.get(0)));
			newList.add(totalize(lists.get(1)));
			return totalize(newList);
		}
	}

	public static void clear() {
		expressions = null;
		newVars = null;
		Expression.clear();
		Runtime.getRuntime().gc();
	}

}
