package iSugar;

import static iSugar.Solver.Status.SATISFIABLE;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.csp.BooleanVariable;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.encoder.Encoder;

/**
 * 
 * @author sako
 *
 */
public class Enumerator extends Solver{

	public Enumerator(String solverName) {
		super(solverName);
	}

	/**
	 * Enumerate solutions.
	 * @throws SugarException
	 * 
	 * @throws IOException
	 * 
	 */
	public void solveProblem() throws SugarException, IOException{

		// temporal files(for non-incremental search)
		String satFileName = "tmp.cnf";
		String mapFileName = "tmp.map";
		String outFileName = "tmp.out";
		ProcessBuilder pb = new ProcessBuilder(solverName, satFileName, outFileName);
		if(!incremental && solverName.equals("sat4j")){
			System.out.println("c solving with sat4j, automatically switched to incremental search.");
			incremental = true;
		}
		
		int numberOfSolve = 0;
		double solveTime = 0.0; // time of each search
		double solveTotalTime = 0.0; // total time of searches
		double totalTime = 0.0; // time of total (include encoding/decoding, adding blocking clause etc)
		ArrayList<Integer> tmpArray = new ArrayList<Integer>(); // for blocking clause
		boolean isSAT = false; // if one solution is found, turn to be true
	
		if(incremental){
			encode();
			totalTime = getTime();
		} else {
			encoder = new Encoder(csp);
			totalTime = System.nanoTime();
			encoder.encode(satFileName);
			encoder.outputMap(mapFileName);
		}
		
		// searching routine
		while(true){
  			System.out.printf("c %dth Solving...\n", numberOfSolve+1);
  			solveTime = incremental ? getTime() : System.nanoTime();
 
  			// invoke SAT solver (non-incremental search)
  			if(!incremental){
  				pb.redirectErrorStream(true);
  				try {
  					Process p = pb.start();
  					p.waitFor();
  					System.out.println(pb.redirectInput());
  					try (BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()))) {
  						for(String line = br.readLine(); line != null; line = br.readLine()) {
  							System.out.println(line);
  						}
  					}
  				} catch (IOException | InterruptedException e) {
  					e.printStackTrace();
  				}
  			}
  			
  			// solve
  			if(incremental ? solveCSP()==SATISFIABLE : encoder.decode(outFileName)){
  				solveTime = incremental ? getTime()-solveTime : System.nanoTime()-solveTime;
  				solveTotalTime += solveTime;
  				numberOfSolve++;
  				isSAT = true;
  				if(incremental) System.out.println("s SATISFIABLE");
  				
  				// decode(incremental search)
  				if(incremental){
  					BitSet model = getModel();
  					for (IntegerVariable v : csp.getIntegerVariables()) {
  						v.decode(model);
  					}
  					for (BooleanVariable v : csp.getBooleanVariables()) {
  						v.decode(model);
  					}
  				}		
  				
  				// make blocking clause
  				for(IntegerVariable v : csp.getIntegerVariables()){
  					System.out.println("a " + v.getName() + "\t" + v.getValue());
  					if(!v.isAux()){
  						int lit;
  						if(v.getValue()!=v.getDomain().getLowerBound()){
  							lit = v.getCode() + v.getDomain().sizeLE(v.getValue())-2;
  							tmpArray.add(incremental ? -lit : lit);
  						}
  						if(v.getValue()!=v.getDomain().getUpperBound()){
  							lit = v.getCode() + v.getDomain().sizeLE(v.getValue())-1;
  							tmpArray.add(incremental ? lit : -lit);
  						}
  					}
  				}
  				System.out.println("a ");
  				
  				int size = tmpArray.size();						
  				int[] blockingClause = new int[size];					
  				for(int i=0; i<size; i++){
  					blockingClause[i] = tmpArray.get(i);
  				}

  				// add blocking clause
  				if(incremental){
  					addBlockingClause(blockingClause);
  				} else {
  					StringBuilder sb = new StringBuilder("");
  					File file = new File(satFileName);
  					BufferedReader br = new BufferedReader(new FileReader(file));
  					Pattern p = Pattern.compile("(p cnf \\d+ )(\\d+)");
  					Matcher m;
  					String line = br.readLine();
  					m = p.matcher(line);
  					if(m.find()){
  						int clauses = Integer.parseInt(m.group(2)) + 1;
  						sb.append(m.group(1)+clauses+"\n");
  					}
  					while((line=br.readLine()) != null){
  						sb.append(line+"\n");
  					}
  					br.close();
  					FileWriter filewriter = new FileWriter(file);
  					filewriter.write(sb.toString());
  					for(int i: blockingClause){
  						filewriter.write(i+" ");
  					}
  					filewriter.write("0\n");
  					filewriter.close();
  				}
  				tmpArray.clear();
  			} else {
  				// end of search
  				solveTime = incremental ? getTime()-solveTime : System.nanoTime()-solveTime;
  				solveTotalTime += solveTime;
  				if(incremental) System.out.println("s UNSATISFIABLE");
  				System.out.println("c \nc FINISHED");
  				System.out.println(isSAT ? "c FOUND SOLUTION" : "c NOT FOUND SOLUTION");
  				System.out.printf("c Number of Solutions : %d\n", numberOfSolve);
  				System.out.printf("c Solving Time : %f sec\n", incremental ? solveTotalTime : solveTotalTime/1000000000.0);				
  				break;
  			}
		}
		totalTime = incremental ? getTime()-totalTime : System.nanoTime()-totalTime;
		System.out.printf("c Total Time : %f sec\n", incremental ? totalTime : totalTime/1000000000.0);
		new File(satFileName).delete();
		new File(mapFileName).delete();
		new File(outFileName).delete();
	}
	
}
