package iSugar;

import static iSugar.Solver.Status.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jp.kobe_u.sugar.Logger;
import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.converter.Converter;
import jp.kobe_u.sugar.csp.CSP;
import jp.kobe_u.sugar.csp.IntegerDomain;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.encoder.Encoder;
import jp.kobe_u.sugar.expression.Expression;

/**
 * This class is for solving Hamiltonian cycle problem.
 * @author sako
 *
 */
public class HCPSolver extends Solver {

	Graph graph;
	private int countOfIterations = 0;
	private int countOfCycles = 0;
	private double start, stop, encodetime, solvetime;
	private int counter = 0;

	public HCPSolver(String solverName) {
		super(solverName);
		csp = new CSP();
		graph = new Graph();
	}

	// get a cycle from SAT solution
	public List<Integer> getCycle(int node, int initial, List<Integer> cycle){
		int node2 = getNextNode(node);
		cycle.add(0, node2);
		if(node2 != initial){
			cycle = getCycle(node2, initial, cycle);
		}
		return cycle;
	}
	
	// get the next node adjacent to a node
	private int getNextNode(int node) {
		//System.out.println("node : "+node);
		Set<Integer> adjacentNodes = graph.getAdjacent(node);
		for(int n : adjacentNodes){
			//System.out.print("a_"+node+"_"+n + " ");
			int a = csp.getIntegerVariable("a_"+node+"_"+n).getValue();
			if(a>0){
				//System.out.println(a+" hoge");
				return n;
			}
		}
		System.out.println("ERROR");
		System.exit(1);
		return 0;
	}

	// get all cycles
	public Set<List<Integer>> getCycles(){
		Set<List<Integer>> cycles = new LinkedHashSet<List<Integer>>();
		List<Integer> nodes1 = new ArrayList<Integer>(graph.getNodes());
		while(!nodes1.isEmpty()){
			int node = nodes1.get(0);
			List<Integer> cycle = new ArrayList<Integer>();
			cycle.add(node);
			cycle = getCycle(node, node, cycle);
			cycles.add(cycle);
			nodes1.removeAll(cycle);
		}
 		return cycles;
	}

	
	// load the file and add the information to the graph object
	public void parseColFile(String fileName){
		try {
			BufferedReader rd = new BufferedReader(new FileReader(fileName));
			String line;
			String regex = "e\\s+(\\d+)\\s+(\\d+)";
			Pattern p = Pattern.compile(regex);
			Matcher m;
			// ignore except the line "e (num) (num)"
			while(true){
				line = rd.readLine();
				if(line == null) break;
				if(!line.trim().startsWith("e")) continue;
				m = p.matcher(line);
				if(m.find()){
					int n1 = Integer.parseInt(m.group(1));
					graph.addNode(n1);
					int n2 = Integer.parseInt(m.group(2));
					graph.addNode(n2);
					graph.addEdge(n1, n2);
				}	
			}
			rd.close();			
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found.");
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	// generate a CSP
	public void define() throws SugarException {
		IntegerDomain domain = IntegerDomain.create(0, 1);
		String name;
		List<Expression> allExpressions = new ArrayList<Expression>();
		// definition the vertices and constraint of edges
		for(int[] edge : graph.getEdges()){
			List<Expression> expressions = new ArrayList<Expression>();
			List<Expression> innerExpressions = new ArrayList<Expression>();
			expressions.add(Expression.LE); // [<= ]
			innerExpressions.add(Expression.ADD); // [+ ]
			name = "a_"+edge[0]+"_"+edge[1];
			try{
				IntegerVariable v1;
				v1 = new IntegerVariable(name, domain);
				csp.add(v1);
				innerExpressions.add(Expression.create(name)); // [+ a_ij]
				name = "a_"+edge[1]+"_"+edge[0];
				IntegerVariable v2 = new IntegerVariable(name, domain);
				csp.add(v2);
				innerExpressions.add(Expression.create(name)); // [+ a_ij a_ji]
				expressions.add(Expression.create(innerExpressions)); // [<= [+ a_ij a_ji]]
				expressions.add(Expression.create(1)); // [<= [+ a_ij a_ji] 1]  a_ij + a_ji <= 1 
				allExpressions.add(Expression.create(expressions));
			} catch(SugarException e){
				continue;
			}
		}
		// constraint of in and out degree
		int listSize; // >=2
		for(int n1: graph.getNodes()){
			List<Integer> adjacentList = new ArrayList<Integer>(graph.getAdjacent(n1));			
			listSize = adjacentList.size();
			IntegerVariable s;
			List<Expression> expressions_in = new ArrayList<Expression>();
			List<Expression> innerExpressions_in = new ArrayList<Expression>();
			List<Expression> expressions_out = new ArrayList<Expression>();
			List<Expression> innerExpressions_out = new ArrayList<Expression>();
			// UNSAT if size of the list is 1 (cannot generate any cycle)
			if(listSize < 2){
				System.out.println("This Graph has a node that has only zero or one adjacent node");
				System.out.println("This Graph has no Hamiltonian Cycle");
				System.exit(0);
			}
			// size = 2
			else if(listSize == 2){
				// in degree
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(0)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "a_"+adjacentList.get(1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 a_(i+1)n1]			
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 a_(i+1)n1]]
				innerExpressions_in.clear();
				expressions_in.add(Expression.create(1)); // [= [+ a_in1 a_(i+1)n1] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// out degree
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(0);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "a_"+n1+"_"+adjacentList.get(1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i a_n1(i+1)]			
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i a_n1(i+1)]]
				innerExpressions_out.clear();
				expressions_out.add(Expression.create(1)); // [= [+ a_n1i a_n1(i+1)] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();
			}
			// introduce complement variable if size >= 3
			else{
				// in degree
				// s = x1 + x2
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(0)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "a_"+adjacentList.get(1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 a_(i+1)n1]			
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 a_(i+1)n1]]
				innerExpressions_in.clear();
				name = "s_"+counter;
				//counter++;
				s = new IntegerVariable(name, domain);
				csp.add(s);
				expressions_in.add(Expression.create(name)); // [= [+ a_in1 a_(i+1)n1] s]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// s' = x3 + s
				for(int i=2; i<listSize-1; i++){
					expressions_in.add(Expression.EQ); // [= ]
					innerExpressions_in.add(Expression.ADD); // [+ ]			
					name = "a_"+adjacentList.get(i)+"_"+n1;
					innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
					name = "s_"+counter;
					counter++;
					innerExpressions_in.add(Expression.create(name)); // [+ a_in1 s]
					expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 s]]
					innerExpressions_in.clear();
					name = "s_"+counter;
					IntegerVariable s1 = new IntegerVariable(name, domain);
					csp.add(s1);
					expressions_in.add(Expression.create(name)); // [= [+ a_in1 s] s1]
					allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
					expressions_in.clear();
				}	
				// 1 = xn + s''
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(listSize-1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "s_"+counter;
				counter++;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 s]
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 s]]
				innerExpressions_in.clear();
				expressions_in.add(Expression.create(1)); // [= [+ a_in1 s] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// out degree
				// s = x1 + x2
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(0);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "a_"+n1+"_"+adjacentList.get(1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i a_n1(i+1)]			
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i a_n1(i+1)]]
				innerExpressions_out.clear();
				name = "s_"+counter;
				//counter++;
				s = new IntegerVariable(name, domain);
				csp.add(s);
				expressions_out.add(Expression.create(name)); // [= [+ a_n1i a_n1(i+1)] s]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();
				// s' = x3 + s
				for(int i=2; i<listSize-1; i++){
					expressions_out.add(Expression.EQ); // [= ]
					innerExpressions_out.add(Expression.ADD); // [+ ]			
					name = "a_"+n1+"_"+adjacentList.get(i);
					innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
					name = "s_"+counter;
					counter++;
					innerExpressions_out.add(Expression.create(name)); // [+ a_n1i s]
					expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i s]]
					innerExpressions_out.clear();
					name = "s_"+counter;
					IntegerVariable s1 = new IntegerVariable(name, domain);
					csp.add(s1);
					expressions_out.add(Expression.create(name)); // [= [+ a_n1i s] s1]
					allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
					expressions_out.clear();
				}	
				// 1 = xn + s''
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(listSize-1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "s_"+counter;
				counter++;
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i s]
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i s]]
				innerExpressions_out.clear();
				expressions_out.add(Expression.create(1)); // [= [+ a_n1i s] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();
			}
		}
		Converter converter = new Converter(csp);
		converter.convert(allExpressions);
		//csp.output(System.out, ""); // for debug
		converter = null;
		allExpressions = null;
		Expression.clear();
		Runtime.getRuntime().gc();
	}
	
	// add blocking clauses that is not Hamiltonian cycle
	public void addBlockingClauses(List<Integer> cycle) throws SugarException{
		int size = cycle.size();
		//System.out.println("size : "+size);
		int n1, n2;
		int[] blockingClause1 = new int[size-1], blockingClause2 = new int[size-1];
		for(int i = 0; i<size-1; i++){
			n1 = cycle.get(i); n2 = cycle.get(i+1);
			blockingClause1[i] = csp.getIntegerVariable("a_"+n1+"_"+n2).getCode();
			blockingClause2[i] = csp.getIntegerVariable("a_"+n2+"_"+n1).getCode();
		}
		addNormalizedClause(blockingClause1);
		addNormalizedClause(blockingClause2);
	}

	// add blocking clauses that is not Hamiltonian cycle(not use incremental method)
	 public void addBlockingClauses(List<Integer> cycle, String satFileName) throws SugarException, IOException{
		int size = cycle.size();
		//System.out.println("size : "+size);
		int n1, n2;
		ArrayList<Integer> blockingClause1 = new ArrayList<Integer>(), blockingClause2 = new ArrayList<Integer>();
		for(int i = 0; i<size-1; i++){
			n1 = cycle.get(i); n2 = cycle.get(i+1);
			blockingClause1.add(csp.getIntegerVariable("a_"+n1+"_"+n2).getCode());
			blockingClause2.add(csp.getIntegerVariable("a_"+n2+"_"+n1).getCode());
		}
		// write to csp file
		StringBuilder sb = new StringBuilder("");
		File file = new File(satFileName);
		BufferedReader br = new BufferedReader(new FileReader(file));
		Pattern p = Pattern.compile("(p cnf \\d+ )(\\d+)");
		Matcher m;
			 String line = br.readLine();
			 m = p.matcher(line);
		if(m.find()){
			 int clauses = Integer.parseInt(m.group(2)) + 2;
			 sb.append(m.group(1)+clauses+"\n");
		}
		while((line=br.readLine()) != null){
			 sb.append(line+"\n");
		}
		br.close();
		FileWriter filewriter = new FileWriter(file);
		filewriter.write(sb.toString());
		for(int i: blockingClause1){
			 filewriter.write(i+" ");
		}
		filewriter.write("0\n");
		for(int i: blockingClause2){
			 filewriter.write(i+" ");
		}
		filewriter.write("0\n");
		filewriter.close();
	}
	
	// solve Hamiltonian cycle problem
	public void solveHCP() throws SugarException{
		start = System.nanoTime();
		encode();
		start = System.nanoTime();
		System.out.println("c Searching Hamiltonian Cycle...");
		while(solveCSP()==SATISFIABLE){
			System.out.println("s SATISFIABLE");
			countOfIterations++;
			/*
			for (IntegerVariable v : csp.getIntegerVariables()) {
				System.out.println("code of "+v.getName()+" = "+v.getCode());
			}
			satSolver.outputModel();
			*/
			BitSet model = satSolver.getModel();
			for (IntegerVariable v : csp.getIntegerVariables()) {
				v.decode(model);
				//System.out.println(v.getName()+" = "+v.getValue());
			}
			// get cycles
			Set<List<Integer>> cycles = getCycles();
			if(cycles.size()==1){ // if only one cycle, solution found
				countOfCycles++;
				System.out.println("c A Hamiltonian Cycle is Found");
				System.out.print("a ");
				for(List<Integer> cycle : cycles){
					for(int node : cycle){
						System.out.print(node+" ");
					}
				}
				System.out.println();
				System.out.println("c Iterations : " + countOfIterations);
				System.out.println("c Cycles : "+ countOfCycles);
				stop = System.nanoTime();
				solvetime = (stop-start) / 1000000000.0;
				System.out.printf("c Solving Time : %f sec\n", solvetime);
				System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
				System.exit(0);
			}
			else{ // if more than one cycle, not solution
				System.out.println("c found " + cycles.size() + " cycles");
				countOfCycles += cycles.size();
				for(List<Integer> cycle : cycles)
					addBlockingClauses(cycle);
			}
		}
		System.out.println("c This Graph has no Hamiltonian Cycle");
		System.out.println("c Iterations : " + countOfIterations);
		System.out.println("c Cycles : "+ countOfCycles);
		stop = System.nanoTime();
		solvetime = (stop-start) / 1000000000.0;
		System.out.printf("c Solving Time : %f sec\n", solvetime);
		System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
	}

	 // solve Hamiltonian cycle problem
	 public void solveHCP(String solverName) throws SugarException, IOException {
		String satFileName = "tmp.cnf";
		String mapFileName = "tmp.map";
		String outFileName = "tmp.out";
		start = System.nanoTime();
		Encoder encoder = new Encoder(csp);
		encoder.encode(satFileName);
		encoder.outputMap(mapFileName);
		stop = System.nanoTime();
		encodetime = (stop-start) / 1000000000.0;
		System.out.println("c Finished Encoding.\n");
		System.out.printf("c Encoding Time : %f sec\n", encodetime);
		//System.out.printf("Encode Time : %f sec\n", SatSolver4JNA.addclausetime);
		System.out.printf("c variables : %d\nc clauses : %d\n", variablesCount, clausesCount);

		start = System.nanoTime();
		System.out.println("c Searching Hamiltonian Cycle...");
		// invoke a SAT solver as an another process
		ProcessBuilder pb = new ProcessBuilder(solverName, satFileName, outFileName);
		while(true){
			countOfIterations++;
			pb.redirectErrorStream(true);
			try {
				 Process p = pb.start();	
				 p.waitFor();
				 System.out.println(pb.redirectInput());
				 try (BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()))) {
					 for(String line = br.readLine(); line != null; line = br.readLine())
						 System.out.println(line);
				 }
			} catch (IOException | InterruptedException e) {
				 e.printStackTrace();
			}

			// decode
			if(encoder.decode(outFileName)){
				 // get cycles
				 Set<List<Integer>> cycles = getCycles();
				 if(cycles.size()==1){
					  countOfCycles++;
					  System.out.println("c A Hamiltonian Cycle is Found");
					  System.out.print("a ");
					  for(List<Integer> cycle : cycles){
							for(int node : cycle){
								 System.out.print(node+" ");
							}
					  }
					  System.out.println();
					  System.out.println("c Iterations : " + countOfIterations);
					  System.out.println("c Cycles : "+ countOfCycles);
					  stop = System.nanoTime();
					  solvetime = (stop-start) / 1000000000.0;
					  System.out.printf("c Solving Time : %f sec\n", solvetime);
					  System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
					  System.exit(0);
				 }
				 else{
					  System.out.println("c found " + cycles.size() + " cycles");
					  countOfCycles += cycles.size();
					  for(List<Integer> cycle : cycles)
							addBlockingClauses(cycle, satFileName);
				 }
			} else break;
		}
		System.out.println("c This Graph has no Hamiltonian Cycle");
		System.out.println("c Iterations : " + countOfIterations);
		System.out.println("c Cycles : "+ countOfCycles);
		stop = System.nanoTime();
		solvetime = (stop-start) / 1000000000.0;
		System.out.printf("c Solving Time : %f sec\n", solvetime);
		System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
	 }

	
	/**
	 * Solves HCP with CounterExample-Guided Abstraction Refinement (CEGAR).
	 * @throws SugarException
	 * 
	 * @throws IOException
	 * 
	 */
	public void solveProblem() throws SugarException, IOException {
		System.out.println("c Parsing Problem...");
		parseColFile(problemFile);
		define();
		System.out.println("c Finished.");
        Logger.info(graph.getNodes().size() + " Nodes, " + graph.getEdges().size() + " Edges");
		if (incremental) {
			 solveHCP();
		} else {
			if(solverName.equals("sat4j")){
				System.out.println("c solving with sat4j, automatically switched to incremental search.");
				incremental = true;
				solveHCP();
			}
			 solveHCP(solverName);
		}
	}
	
}
