package iSugar;

import java.io.IOException;

import jp.kobe_u.sugar.Logger;
import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.csp.CSP.Objective;

public class Main {

	private final String version = "iSugar v1.1.0";
	private Solver solver;
	private String solverName = "glueminisat";
	private Type type = Type.CSP;
	
	// problem type
	private enum Type {
		CSP,
		MAXCSP,
		ENUM,
		HCP
	}

	/**
	 * print version of iSugar and back-end solver.
	 */
	public void printSignature(){
		Logger.println("c " + version + " + " + solverName);
	}
	
	public void usage(){
		System.out.println("Usage: java -cp ..:../../jna-4.2.2.jar:../../sugar-v2-2-1.jar iSugar.Main [options] problemfile");
		System.out.println("\t-h -help             : show help");
		System.out.println("\t-v -verbose          : verbose output");
		System.out.println("\t-vv -veryverbose     : very verbose output");
		System.out.println("\t-n                   : non incremental mode");
		System.out.println("\t-all                 : enumeration mode");
		System.out.println("\t-hcp                 : solving Hamiltonian cycle problem");
		System.out.println("\t-max                 : Max-CSP mode");
		System.out.println("\t-solver command      : SAT solver command (default: glueminisat)");
		System.out.println("\t-encoding command    : encoding mode of BC constraint command (default: totalizer)");
		System.out.println("\t\t-totalizer         : totalizer method");
		System.out.println("\t\t-seqcounter        : sequential counter method");
		System.out.println("\t-strategy command    : optimization strategy command (default: binary search)");
		System.out.println("\t\t-bin               : binary search");
		System.out.println("\t\t-mix               : mixed search");
		System.out.println("\t\t-us0 -v0           : unsat based search (version 0)");
		System.out.println("\t\t-us1 -v1           : unsat based search (version 1)");
		System.out.println("\t\t-us2 -v2[,c1][,c2] : unsat based search (version 2)");
		System.out.println("\t\t-us3 -v3[,c1][,c2] : unsat based search (version 3)");
		System.out.println("\t\t\t,n               : non-assumption size: n");
		System.out.println("\t\t\t,v               : non-assumption size: v/n (v: domain of objective variable in each solving)");
		System.out.println("\t\t\t,2d              : non-assumption size: a half of domain of objvar");
		System.out.println("\t\t\t,4d              : non-assumption size: a quarter of domain of objvar");
		System.out.println("\t-threshold command   : optimization threshold command (default: 100)");
		System.out.println("\t\t-n                 : threshold = n");
		System.out.println("\t\t-nd                : threshold = n/d (d: initial domain of objective variable)");
		System.exit(0);
	}

	public void launch(){
		try {
			printSignature();
			switch (type){
			case CSP:
				Solver.loadCSP();
				if(Solver.csp.getObjective().equals(Objective.NONE))
					solver = new CSPSolver(solverName);
				else	
					solver = new Optimizer(solverName);
				break;
			case HCP:
				solver = new HCPSolver(solverName);
				break;
			case ENUM:
				Solver.loadCSP();
				solver = new Enumerator(solverName);
				break;
			case MAXCSP:
				Solver.isMaxCSP = true;
				Solver.loadCSP();
				solver = new MaxCSPSolver(solverName);
				break;
			}
			solver.solveProblem();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public void parseOptions(String[] args) throws SugarException{
		int i = 0;
		if(args.length==0) usage();
		while(i < args.length){
			if(args[i].equals("-h") || args[i].equals("-help"))
				usage();
			else if(args[i].equals("-i"))
				Solver.incremental = true;
			else if(args[i].equals("-n"))
				Solver.incremental = false;
			else if(args[i].equals("-all"))
				type = Type.ENUM;
			else if(args[i].equals("-hcp"))
				type = Type.HCP;
			else if(args[i].equals("-v"))
				Logger.verboseLevel = 1;
			else if(args[i].equals("-vv"))
				Logger.verboseLevel = 2;
			else if(args[i].equals("-max"))
				type = Type.MAXCSP;
			else if(args[i].equals("-solver")){
				if(i+1<args.length)
					solverName = args[i+1];
				else
                    throw new SugarException("Solver is missing : -solver");		
			}
			else if(args[i].equals("-encoding")){
				if(i+1<args.length)
					ChannelingConstraintEncoder.encoding = args[i+1];
				else
                    throw new SugarException("Encoding is missing : -encoding");		
			}
			else if(args[i].equals("-strategy")){
				if(i+1<args.length)
					Solver.strategy = args[i+1];
				else
                    throw new SugarException("Strategy is missing : -strategy");		
			}
			else if(args[i].equals("-threshold")){
				if(i+1<args.length)
					Optimizer.threshold = args[i+1];
				else
                    throw new SugarException("Threshold is missing : -threshold");		
			}
			else if(args[i].startsWith("-"))
                throw new SugarException("Unknown Option : " + args[i]);
			else
				Solver.problemFile = args[i];
			i++;		
		}
	}
	
	public static void main(String[] args) throws SugarException, IOException {
		Main main = new Main();
		main.parseOptions(args);
		main.launch();
	}
}
