package iSugar;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.csp.BooleanVariable;
import jp.kobe_u.sugar.csp.CSP;
import jp.kobe_u.sugar.expression.Expression;

/**
 * This class is for solving MaxCSP.
 * @author sako
 *
 */
public class MaxCSP extends CSP {
	
	private List<BooleanVariable> relaxationVariables; 
	private int[] litsOfRelaxationVariables;
	private int[] weights;
	private boolean[] subsetSumOfWeights;
	private List<BooleanVariable> assumedVariables;
	private List<BooleanVariable> notAssumedVariables;
	private int[] litsOfNotAssumedVariables = {};
	private int[] weightsOfNotAssumedVariables = {};
	private List<BooleanVariable> auxiliaryVariables;
	private int sumOfWeights = 0;
	private int cost = 0;
	private int bound = 0;
	private BooleanVariable tempVar;

	public static final String RELAXATION = "_r"; 
	public static final String AUXILIARY = "_a"; 
	public static final boolean GEQ = true;
	public static final boolean LEQ = false;
	public static boolean isWeighted = false;
	public static int numOfSoftConstraints = 0;

	// constructor
	public MaxCSP(){
		super();
		relaxationVariables = new ArrayList<BooleanVariable>();
		assumedVariables = new ArrayList<BooleanVariable>();
		notAssumedVariables = new ArrayList<BooleanVariable>();
		auxiliaryVariables = new ArrayList<BooleanVariable>();
	}

	/**
	 * relax constraints of CSP for MaxCSP (modified SugarMain.toMaxCSP).
	 * 
	 * @param expressions0
	 *  expressions of constraints before relaxation
	 * @return
	 *  expressions of constraints after relaxation
	 * @throws SugarException
	 * 
	 */
    public static List<Expression> relaxConstraint(List<Expression> expressions0) throws SugarException {
        List<Expression> expressions = new ArrayList<Expression>();
        for (Expression x: expressions0) {
            if (x.isSequence(Expression.DOMAIN_DEFINITION)
                || x.isSequence(Expression.INT_DEFINITION)
                || x.isSequence(Expression.BOOL_DEFINITION)
                || x.isSequence(Expression.PREDICATE_DEFINITION)
                || x.isSequence(Expression.RELATION_DEFINITION)) {
                expressions.add(x);
            } else if (x.isSequence(Expression.OBJECTIVE_DEFINITION)) {
                throw new SugarException("Illegal " + x);
            } else {
                Expression r = Expression.create(RELAXATION + numOfSoftConstraints);
                expressions.add(Expression.create(Expression.BOOL_DEFINITION, r));
                x = r.or(x);
                expressions.add(x);
                numOfSoftConstraints++;
            }
        }
        return expressions;
    }

	/**
	 * return the sum of weights of each soft clauses.
	 * 
	 * @return sum of weights
	 */
	public int getSumOfWeight(){
		return sumOfWeights;
	}
	
	/**
	 * make the array of subset sum of weights.
	 * each index of this array represents subset sum.
	 */
	private void makeSubsetSumOfWeights(){
		subsetSumOfWeights = new boolean[sumOfWeights+1];
		subsetSumOfWeights[0] = true;
		for(int i=0; i<weights.length; i++)
			for(int j=sumOfWeights; j>=weights[i]; j--)
				if(subsetSumOfWeights[j-weights[i]])
					subsetSumOfWeights[j] = true;
	}
	
	/**
	 * relax soft constraints of CSP.
	 * each soft constraint "C" is relaxed to "r or C" using a relaxation variable "r".
	 * and calculate the sum of weight.
	 */
	public void registerRelaxationVariables(){
		for (BooleanVariable b : this.getBooleanVariables()){
			if(b.getName().startsWith(RELAXATION))
				relaxationVariables.add(b);
		}
		int n = relaxationVariables.size();
		litsOfRelaxationVariables = new int[n];
		weights = new int[n];
		for (int i=0; i<n; i++){
			litsOfRelaxationVariables[i] = relaxationVariables.get(i).getCode();
			weights[i] = 1;
			sumOfWeights += weights[i];
		}
		if(isWeighted)
			makeSubsetSumOfWeights();
		assumedVariables = new ArrayList<BooleanVariable>(relaxationVariables);
	}

	/**
	 * gets assumption literals.
	 * 
	 * @return 
	 * 	array of assumption literals
	 */
	public int[] getAssumptions(){
		int[] assumptions = new int[assumedVariables.size()];
		int i = 0;
		for(BooleanVariable a : assumedVariables)
			assumptions[i++] = -a.getCode();
		return assumptions;
	}
	
	/**
	 * add constraint [wi * ri <= k] to CSP.
	 * 
	 * @param k
	 *  bound
	 */
	public void addSumOfWeightsConstraint(Solver solver, int k){
        solver.addPB(weights, litsOfRelaxationVariables, k, LEQ);
	}

	/**
	 * get bound.
	 * 
	 * @return 
	 *  bound
	 */
	public int getBound(){
		return bound;
	}
	
	/**
	 * update literals of not-assumed variables.
	 */
	public void updateLitsOfNotAssumedVariables(){
        int nasize = notAssumedVariables.size();
        int asize = litsOfNotAssumedVariables.length;
        int newLitsOfNotAssumedVariables[] = new int[nasize];
        int newWeightsOfNotAssumedVariables[] = new int[nasize];
        System.arraycopy(litsOfNotAssumedVariables, 0, newLitsOfNotAssumedVariables, 0, asize);
        System.arraycopy(weightsOfNotAssumedVariables, 0, newWeightsOfNotAssumedVariables, 0, asize);
        for(int i=asize; i<nasize; i++){
        	newLitsOfNotAssumedVariables[i] = notAssumedVariables.get(i).getCode();
        	newWeightsOfNotAssumedVariables[i] = weights[relaxationVariables.indexOf(notAssumedVariables.get(i))];
        	bound += newWeightsOfNotAssumedVariables[i];
        }
        litsOfNotAssumedVariables = newLitsOfNotAssumedVariables;
        weightsOfNotAssumedVariables = newWeightsOfNotAssumedVariables;
	}
	
	/**
	 * remove constraint used an assumption literal from SAT solver.
	 * the constraint "¬a or C" is satisfied by assigning false to the assumption literal "a".
	 * 
	 * @param solver
	 *  solver 
	 */
	public void removeTempSumOfWeightsConstraint(Solver solver){
		int bc[] = {-tempVar.getCode()};
		solver.add_clause(bc);
	}
	
	/**
	 * add constraint "wi * ri <= k" to CSP.
	 * each consraint "C" is added as "¬a or C" using a auxiliary variable "a", and "a" is assumed true. 
	 * the added constraint by this method need to removed after next solving,
	 * by calling "removeTempSumOfWeightsConstraint()".
	 * 
	 * @param solver
	 *  solver object
	 * @param k
	 *  bound
	 * @throws SugarException 
	 * 
	 * @throws IOException 
	 * 
	 */
	public void addTempSumOfWeightsConstraint(Solver solver, int k) throws SugarException, IOException {
		if(tempVar!=null)
			removeTempSumOfWeightsConstraint(solver);
		updateLitsOfNotAssumedVariables();
		if(litsOfNotAssumedVariables.length!=0 && bound > k){
		//if(litsOfNotAssumedVariables.length!=0){
			int n = auxiliaryVariables.size();
			tempVar = new BooleanVariable(AUXILIARY + n);
			this.add(tempVar);
			solver.encodeDelta();
			this.commit();
			//System.out.println("code of aux : " + tempVar.getCode());
			auxiliaryVariables.add(tempVar);
			//System.out.println("c adds constraint : " + tempVar.getName() + " => (wi*ri <= " + k + ")");
			System.out.println("c ADD TEMP CONSTRAINT : " + tempVar.getName() + " => (FALSE CONSTRAINTS <= " + k +")");
			solver.addPB(weightsOfNotAssumedVariables, litsOfNotAssumedVariables, k, LEQ, tempVar.getCode());
		}
	}
	
	/**
	 * return a cost of latest model.
	 * the cost is sum of weights of constraint unsatisfied.
	 * 
	 * @return
	 *  cost
	 */
	public int getCost(){
		cost = 0;
		// not precise...
		for(int i=0; i<relaxationVariables.size(); i++)
			if(relaxationVariables.get(i).getValue()==true)
				cost += weights[i];
		return cost;
	}
	
	/**
	 * update upper bound of search.
	 * 
	 * @param ub
	 *  old upper bound
	 * @return
	 *  new upper bound
	 */
	public int refineUpperBound(int ub){
		int n = 0;
		// not precise...
		for(BooleanVariable r : relaxationVariables)
			if(r.getValue())
				n += weights[relaxationVariables.indexOf(r)];
		return n;
		//return --lb;
	}
	
	/**
	 * update lower bound of search.
	 * 
	 * @param lb
	 *  old lower bound
	 * @return
	 *  new lower bound
	 */
	public int refineLowerBound(int lb){
		if(isWeighted)
			return subsetSum(lb);
		return ++lb; 
	}
	
	/**
	 * 
	 * 
	 * @param bound
	 *  previous bound
	 * @return
	 *  minimum subset sum that is greater than bound
	 */
	private int subsetSum(int bound){
		for(int i=bound+1; i<subsetSumOfWeights.length; i++){
			if(subsetSumOfWeights[i])
				return i;
		}
		return ++bound;
	}
	
	/**
	 * remove assumption literals used for proving unsatisfiability.
	 * if any removed literal is nothing, update upper bound of search.
	 * 
	 * @param literals
	 *  result of assumption literals
	 * @return
	 *  true if removed at least one literal, false otherwise
	 */
	public boolean updateAssumptionLiterals(int[] literals){
		int removed = 0;
		for (int i=literals.length-1; i>=0; i--){
			if(literals[i]==1){
				notAssumedVariables.add(assumedVariables.remove(i));
				removed++;
			}
		}
		if(removed != 0){
			System.out.println("c RELAXED " + removed + " CONSTRAINTS (REST : "+ (literals.length-removed) + ")");
			//for (BooleanVariable b : assumedVariables)
				//System.out.print(b.getName() + " ");
			//System.out.println(")");
			return true;
		}
		return false;
		//return literals.length != assumedVariables.size();
	}
	
}
