package iSugar;

import static iSugar.Solver.Status.*;
import iSATLibrary.SatSolverForNative;
import iSATLibrary.SatSolverForSat4j;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jp.kobe_u.sugar.Logger;
import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.csp.CSP;
import jp.kobe_u.sugar.csp.IntegerDomain;
import jp.kobe_u.sugar.csp.IntegerVariable;

/**
 * This class is for solving optimization problem.
 * @author sako
 *
 */
public class Optimizer extends Solver {
	
	private int numberOfSolve = 0;
	private double solveTime = 0.0, optSolveTime = 0.0;
	private int res;

	private IntegerVariable objVar;
	private IntegerDomain domOfObjVar;
	private ArrayList<Integer> domValOfObjVar;
	private Iterator<Integer> iter;
	
	private List<IntegerVariable> vList = new ArrayList<IntegerVariable>();
	
	private int assumptionMode = 0;
	private boolean isFixed = true;
	
	public static String threshold = "100";
	
	public Optimizer(String solverName) {
		super(solverName);
	}
	
	public int setThreshold(int domSize){
		int th = 100;
		try{
			th = Integer.parseInt(threshold);
		} catch(NumberFormatException e){
			if(threshold.equals("2d"))
				th = domSize/2;
			else if(threshold.equals("4d"))
				th = domSize/4;
			else if(threshold.equals("8d"))
				th = domSize/8;
		}
		return th;
	}
	
	public void setAssumptionMode(){
		String[] s = strategy.split(",");
		int i=1;
		while(i < s.length){
			try {
				assumptionMode = Integer.parseInt(s[i]);
			} catch(NumberFormatException e){
				if(s[i].equals("v"))
					isFixed = false;
			}
			i++;
		}
	}
	
	public void setup() throws SugarException{
		encode();

		objVar = csp.getObjectiveVariables().get(0); // objective variable
		domOfObjVar = objVar.getDomain();
		domValOfObjVar= new ArrayList<Integer>();
		
		iter = domOfObjVar.values();
		while (iter.hasNext()) { domValOfObjVar.add(iter.next()); }
	}

	/**
	 * solves an optimization problem by SAT based search.
	 * binary and mixed search are available.
	 * @throws SugarException
	 * 
	 */
	public void optimize() throws SugarException {
		
		boolean isMin = csp.getObjective().equals(CSP.Objective.MINIMIZE);
		boolean isMix = strategy.equals("mix");

		setup();

		int assumption[] = new int[1];

		int mid; // the value of assumption (p(x<=mid) or -p(x<=mid))
		int offset = 0; 
		boolean isSAT = false; // if one solution is found, turn to be true
		
		// searching routine
		optSolveTime = getTime();
		while (true) {
			
			// making assumption clause
			if(isMix && domValOfObjVar.size() < 100){
				mid = isMin ? domValOfObjVar.size()-1 : 0;
			} else {
				mid = (isMin || domValOfObjVar.size()==1) ? domValOfObjVar.size() / 2 : domValOfObjVar.size() / 2 - 1;
			}
			assumption[0] = (isMin || domValOfObjVar.size()==1) ? objVar.getCode() + mid + offset : -(objVar.getCode() + mid + offset);

			System.out.println("c " + ++numberOfSolve + "th Solving.");
			System.out.println("c OBJECTIVE BOUND: " + domValOfObjVar.get(0) + " <= " + objVar.getName()
					+ " <= " + domValOfObjVar.get(domValOfObjVar.size() - 1));
			System.out.println("c SEARCHING: " + objVar.getName()
					+ ((isMin || domValOfObjVar.size()==1) ? " <= " : " > ") + domValOfObjVar.get(mid) + " ...");

			// solving
			solveTime = getTime();
			// if assumption[0] equals to upper bound, do not add assumption
			res = (isMin && assumption[0]==objVar.getCode()+domOfObjVar.size()-1) ? solveCSP() : solveCSP(assumption);
			if (res==SATISFIABLE) { // SAT
				isSAT = true;
				solveTime = getTime() - solveTime;
				System.out.println("c FOUND: "	+ objVar.getName()
						+ ((isMin || domValOfObjVar.size()==1) ? " <= " : " > ") + domValOfObjVar.get(mid));
				System.out.println("c SOLVING: " + solveTime + " s");
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("o " + objVar.getValue());		
				outputSolution();

				// update domain
				int val = mid;
				if(isMin){
					for (int i = 0; i < mid; i++) {
						if (domValOfObjVar.get(i) == objVar.getValue()) {
							val = i;
							break;
						}
					}
					while (domValOfObjVar.size() > val) { domValOfObjVar.remove(val); }
				} else {
					if (domValOfObjVar.size()==1) break;
					for (int i = mid+1; i < domValOfObjVar.size(); i++) {
						if (domValOfObjVar.get(i) == objVar.getValue()) {
							val = i;
							break;
						}
					}
					for (int i = 0; i < val; i++){
						domValOfObjVar.remove(0);
						offset++;
					}
				}
				// end of search
				if(isMin ? domValOfObjVar.size()==0 : domValOfObjVar.size()==1){
					optSolveTime = getTime()-optSolveTime;
					System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
					System.out.println("s OPTIMUM FOUND");
					System.out.println("c Number of Solves : " + numberOfSolve);
					System.out.println("c Total Solving Time : " + optSolveTime);
					return;
				}
				
				// add clause to SAT solver as fact
					assumption[0] = isMin ? objVar.getCode() + val - 1 + offset : -(objVar.getCode() + offset - 1);
					add_clause(assumption);
			} else { // UNSAT
				solveTime = getTime() - solveTime;
				System.out.println("c NOT FOUND: "	+ objVar.getName()
						+ ((isMin || domValOfObjVar.size()==1) ? " <= " : " > ") + domValOfObjVar.get(mid));
				System.out.println("c SOLVING: " + solveTime + " s");

				// update domain
				if(isMin){
					for (int i = 0; i <= mid; i++) {
						domValOfObjVar.remove(0);
						offset++;
					}
				} else {
					while (domValOfObjVar.size() > mid+1) { domValOfObjVar.remove(mid+1); }
				}
				// end of search
				if (isMin ? domValOfObjVar.size() == 0 : domValOfObjVar.size() == 1) {
					optSolveTime = getTime()-optSolveTime;
					if (isSAT){
						System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
						System.out.println("s OPTIMUM FOUND");
					}
					else
						System.out.println("s UNSATISFIABLE");
					System.out.println("c Number of Solves : " + numberOfSolve);
					System.out.println("c Total Solving Time : " + optSolveTime + " s");
					return;
				}
				
				// add clause to SAT solver as fact
				assumption[0] *= -1;
				add_clause(assumption);
			}
		}
	}

	/**
	 * solves an optimization problem by unsat based algorithm.
	 * @throws SugarException
	 * 
	 */
	public void optimize_us_v0() throws SugarException{
		
		setup();
		
		int lbCode = objVar.getCode();
		int ubCode = lbCode + domOfObjVar.size() - 2;		

		// searching routine
		optSolveTime = getTime();
		while (true) {

			System.out.println("c " + ++numberOfSolve + "th Solving.");
			int lb = lbCode-objVar.getCode();
			int ub = ubCode-objVar.getCode()+1;
			System.out.println("c OBJECTIVE BOUND: " + domValOfObjVar.get(lb) + " <= " + objVar.getName()
					+ " <= " + domValOfObjVar.get(ub));
			System.out.println("c SEARCHING: " + objVar.getName() + " <= " + domValOfObjVar.get(lb));

			// assume all codes (>=lb) 
			int[] assumptions = new int[ubCode-lbCode+1];
			//for(int i=0; i<assumptions.length; i++)
			for(int i=assumptions.length-1; i>=0; i--)
				assumptions[i] = lbCode+i;			
			
			// solving
			solveTime = getTime();
			res = solveCSP(assumptions);
			if (res==SATISFIABLE) { // SAT
				solveTime = getTime() - solveTime;
				System.out.println("c FOUND: " + objVar.getName() + " <= " + domValOfObjVar.get(lb));
				System.out.println("c SOLVING: " + solveTime + " s");
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("o " + objVar.getValue());		
				outputSolution();

				optSolveTime = getTime() - optSolveTime;
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("s OPTIMUM FOUND");
				System.out.println("c Number of Solves : " + numberOfSolve);
				System.out.println("c Total Solving Time : " + optSolveTime + " s");
				return;						
			} else { // UNSAT
				solveTime = getTime() - solveTime;
				System.out.println("c NOT FOUND: " + objVar.getName()
						+ " <= " + domValOfObjVar.get(lb));
				System.out.println("c SOLVING: " + solveTime + " s");

				System.out.println("c Extracting assumptions causing UNSAT.");
				int resAssumption[] = satSolver.getFailed(assumptions);
				int smallestCode = -1;
				for(int i=0; i<resAssumption.length; i++){
					if(resAssumption[i]==1){
						smallestCode = i;
						break;
					}
				}
				if(smallestCode >= 0){
					int newLbCode = lbCode+smallestCode+1;
					System.out.println("c REFINE LOWERBOUND");
					lbCode = newLbCode;
				} else {
					optSolveTime = getTime() - optSolveTime;
					System.out.println("s UNSATISFIABLE");
					System.out.println("c Number of Solves : " + numberOfSolve);
					System.out.println("c Total Solving Time : " + optSolveTime + " s");
					return;
				}
			}
		}
	}

	/**
	 * solves an optimization problem by minimal unsat core.
	 * @throws SugarException
	 * @throws IOException
	 * 
	 */
	public void optimize_us_v1() throws SugarException, IOException{
		
		if(satSolver instanceof SatSolverForNative){
			System.out.println("c info: strategy -us1 is available from sat4j only.");
			System.out.println("c       switched to strategy -us2.");
			optimize_us_v2();
			return;
		}
		
		setup();
		
		optSolveTime = System.nanoTime() / 1000000000.0;
		
		int lbCode = objVar.getCode();
		//int ubCode = lbCode + domOfObjVar.size() - 2;
		addChannelingConstraint_forv2(objVar);
		
		int i = 0;
		int[] assumptions = new int[vList.size()];
		for(IntegerVariable v: vList){
			assumptions[i++] = v.getCode();
		}
		SatSolverForSat4j.isParamClause = true;
		SatSolverForSat4j.cIndex++;
		for(int j : assumptions){
			int[] k = {j};
			satSolver.addClause(k);
		}
		SatSolverForSat4j.isParamClause = false;
		
		System.out.println("c " + ++numberOfSolve + "th Solving.");
		System.out.println("c OBJECTIVE BOUND: " + domOfObjVar.getLowerBound() + " <= " + objVar.getName()
				+ " <= " + domOfObjVar.getUpperBound());
		System.out.println("c SEARCHING: " + objVar.getName() + " <= " + domOfObjVar.getLowerBound());
		
		solveTime = getTime();
		res = solveCSP();
		solveTime = getTime() - solveTime;
		if(res == SATISFIABLE){
			System.out.println("c FOUND: "	+ objVar.getName() + " <= " + domOfObjVar.getLowerBound());
			System.out.println("c SOLVING: " + solveTime + " s");
			System.out.println("s SATISFIABLE");
			outputSolution();
			optSolveTime = System.nanoTime() / 1000000000.0 - optSolveTime;
			System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
			System.out.println("s OPTIMUM FOUND");
			System.out.println("c Number of Solves : " + numberOfSolve);
			System.out.println("c Total Solving Time : " + optSolveTime + " s");
		} else {
			System.out.println("c NOT FOUND: "+ objVar.getName() + " <= " + domOfObjVar.getLowerBound());
			System.out.println("c SOLVING: " + solveTime + " s");
			System.out.println("c CALCULATING MINIMAL UNSAT CORE...");
			long start = 0, stop = 0;
			start = System.nanoTime();
			int[] core = satSolver.getMUS();
			stop = System.nanoTime();
			System.out.println("c SIZE OF CORE: " + core.length);
			System.out.println("c CALCULATING: " + (stop-start) / 1000000000.0 + " sec");
			System.out.println("c NOT FOUND: " + objVar.getName() + " <= " + (domOfObjVar.getUpperBound()-core.length));
			
			satSolver.release();
			satSolver = satSolver.init();
			csp.cancel();
			encode();
			
			int codelen = domOfObjVar.size() - 1;
			int codesOfUnsat = codelen - core.length + 1;
			int optcode = lbCode + codesOfUnsat;
			int[] c1 = {optcode};
			int[] c2 = {-(optcode-1)};
			satSolver.addClause(c1);
			satSolver.addClause(c2);
			System.out.println("c REFINE LOWERBOUND");
			
			int optval = domOfObjVar.getLowerBound()+codesOfUnsat;
			System.out.println("c " + ++numberOfSolve + "th Solving.");
			System.out.println("c OBJECTIVE BOUND: " + optval + " <= " + objVar.getName()
					+ " <= " + optval);
			System.out.println("c SEARCHING: " + objVar.getName() + " <= " + optval);
			
			solveTime = getTime();
			res = solveCSP();
			solveTime = getTime() - solveTime;
			if(res == SATISFIABLE){
				System.out.println("s SATISFIABLE");
				outputSolution();
				optSolveTime = System.nanoTime() / 1000000000.0 - optSolveTime;
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("s OPTIMUM FOUND");
				System.out.println("c Number of Solves : " + numberOfSolve);
				System.out.println("c Total Solving Time : " + optSolveTime + " s");
			} else {
				System.out.println("u UNSATISFIABLE");
				optSolveTime = System.nanoTime() / 1000000000.0 - optSolveTime;
				System.out.println("c Number of Solves : " + numberOfSolve);
				System.out.println("c Total Solving Time : " + optSolveTime + " s");
			}

		}
	}

	
	/**
	 * solving optimization problem by sat-unsat based algorithm.
	 * @throws SugarException
	 * @throws IOException
	 */
	public void optimize_us_v2() throws SugarException, IOException{
		
		setup();
		
		final int d = domOfObjVar.size();
		final int lbCode = objVar.getCode();
		//final int ubCode = lbCode + d - 2;
		int[] assumptions = {};
		int border;
		// number of literals not assuming (p_i=0)
		int nonAssumedLits = 0;
		// threshold of switching search mode 
		int threshold = setThreshold(d);
		int divisor = 0;
		if(isFixed)
			nonAssumedLits = assumptionMode;
		else
			divisor = assumptionMode;
		// search mode
		boolean usmode = true;
		addChannelingConstraint_forv2(objVar);
		// number of auxiliary variables remained non-zero
		int restVars = vList.size();
		// offset deciding index of assuming literals
		int offset = 0;
		boolean isSAT = false;
		
		int lb = 0;
		int ub = d-1;
		
		// searching routine
		optSolveTime = getTime();
		while(lb<=ub){
			if(!isFixed)
				nonAssumedLits = (ub-lb)/divisor;
			usmode = ub-lb >= threshold;

			// if domain size >= threshold, search mode is "using UNSAT information"
			if(usmode){
				border = lb+nonAssumedLits;
				int assumptionSize = ub-border;
				if(assumptionSize<=0){
					usmode = false;
					border = ub;
				} else {
					assumptions = new int[assumptionSize];
					for(int i=0; i<assumptionSize; i++)
						assumptions[i] = vList.get(i+lb+offset).getCode();
				}
			}
			// if threshold < 100, switch linear search
			else
				border = ub;

			System.out.println("c " + ++numberOfSolve + "th Solving.");
			System.out.println("c OBJECTIVE BOUND: " + domValOfObjVar.get(lb) + " <= " + objVar.getName()
					+ " <= " + domValOfObjVar.get(ub));
			System.out.println("c SEARCHING: " + objVar.getName() + " <= " + domValOfObjVar.get(border));
			
			solveTime = getTime();
			res = usmode ? solveCSP(assumptions) : solveCSP();
			if(res==SATISFIABLE){ // SAT
				isSAT = true;
				solveTime = getTime() - solveTime;
				System.out.println("c FOUND: "	+ objVar.getName() + " <= " + domValOfObjVar.get(border));
				System.out.println("c SOLVING: " + solveTime + " s");
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("o " + objVar.getValue());		
				outputSolution();
				int margin = 0;
				for(int i=lbCode+lb; i<=lbCode+ub; i++){
					if(satSolver.val(i)>0){
						margin = lbCode+border-i;
						if(i!=lbCode+lb){
							int[] fact = {i-1};
							satSolver.addClause(fact);
						}
						ub = i-lbCode-1;
						break;
					} else if(i==lbCode+ub){
						int[] fact = {i-1};
						satSolver.addClause(fact);
						ub--;
					}
				}
				if(usmode){
					for(int i:assumptions){
						int[] fact = {i};
						satSolver.addClause(fact);
						restVars--;
						offset++;
					}
					int a = assumptions[assumptions.length-1];
					for(int i=1; i<= margin+1; i++){
						int[] fact = {a+i};
						satSolver.addClause(fact);
						restVars--;
						offset++;
					}
				}
				System.out.println("c REFINE UPPERBOUND");
			} else { // UNSAT
				solveTime = getTime() - solveTime;
				System.out.println("c NOT FOUND: "+ objVar.getName() + " <= " + domValOfObjVar.get(border));
				System.out.println("c SOLVING: " + solveTime + " s");
				if(usmode){
					int[] cause = satSolver.getFailed(assumptions);
					int count = 0;
					for(int i: cause){
						if(i!=0)
							count++;
					}
					if(count==0) break;
					int zeroVars = restVars-count+1;
					int start = vList.size()-restVars;
					for(int i=start; i<start+zeroVars; i++){
						int[] fact = {-(vList.get(i).getCode())};
						satSolver.addClause(fact);
					}
					restVars -= zeroVars;
					int[] fact = {-(lbCode+lb+nonAssumedLits)};
					satSolver.addClause(fact);
					System.out.println("c REFINE LOWERBOUND");
					lb += zeroVars;
				} else
					break;
			}
		}
		
		if(isSAT){
			optSolveTime = getTime() - optSolveTime;
			System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
			System.out.println("s OPTIMUM FOUND");
			System.out.println("c Number of Solves : " + numberOfSolve);
			System.out.println("c Total Solving Time : " + optSolveTime + " s");
		} else {
			optSolveTime = getTime() - optSolveTime;
			System.out.println("s UNSATISFIABLE");
			System.out.println("c Number of Solves : " + numberOfSolve);
			System.out.println("c Total Solving Time : " + optSolveTime + " s");
		}
	}
	
	/**
	 * solving optimization problem by sat-unsat based algorithm.
	 * @throws SugarException
	 * @throws IOException
	 */
	public void optimize_us_v3() throws SugarException, IOException{
				
		setup();
		
		final int d = domOfObjVar.size();
		final int lbCode = objVar.getCode();
		//final int ubCode = lbCode + d - 2;
		int[] assumptions = {};
		int border;
		// number of literals not assuming (p_i=0)
		int nonAssumedLits = 0;
		// threshold of switching search mode
		int threshold = setThreshold(d);
		int divisor = 0;
		if(isFixed)
			nonAssumedLits = assumptionMode;
		else
			divisor = assumptionMode;

		// search mode
		boolean usmode = true;
		addChannelingConstraint_forv3(objVar);
		// number of auxiliary variables remained non-zero
		int restVars = vList.size();
		// offset deciding index of assuming literals
		int offset = 0;
		boolean isSAT = false;
		
		int lb = 0;
		int ub = d-1;
		
		// searching routine
		optSolveTime = getTime();
		while(lb<=ub){
			if(!isFixed)
				nonAssumedLits = (ub-lb)/divisor;
			usmode = ub-lb >= threshold;

			// if domain size >= threshold, search mode is "using UNSAT information"
			if(usmode){
				border = lb+nonAssumedLits;
				int assumptionSize = ub-border;
				if(assumptionSize<=0){
					usmode = false;
					border = ub;
				} else {
					assumptions = new int[assumptionSize];
					for(int i=0; i<assumptionSize; i++)
						assumptions[i] = -(vList.get(i+lb+offset).getCode());
				}
			}
			// if threshold < 100, switch linear search
			else
				border = ub;

			System.out.println("c " + ++numberOfSolve + "th Solving.");
			System.out.println("c OBJECTIVE BOUND: " + domValOfObjVar.get(lb) + " <= " + objVar.getName()
					+ " <= " + domValOfObjVar.get(ub));
			System.out.println("c SEARCHING: " + objVar.getName() + " <= " + domValOfObjVar.get(border));
			
			solveTime = getTime();
			res = usmode ? solveCSP(assumptions) : solveCSP();
			if(res==SATISFIABLE){ // SAT
				isSAT = true;
				solveTime = getTime() - solveTime;
				System.out.println("c FOUND: "	+ objVar.getName() + " <= " + domValOfObjVar.get(border));
				System.out.println("c SOLVING: " + solveTime + " s");
				System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
				System.out.println("o " + objVar.getValue());
				outputSolution();
				int margin = 0;
				for(int i=lbCode+lb; i<=lbCode+ub; i++){
					if(satSolver.val(i)>0){
						margin = lbCode+border-i;
						if(i!=lbCode+lb){
							int[] fact = {i-1};
							satSolver.addClause(fact);
						}
						ub = i-lbCode-1;
						break;
					} else if(i==lbCode+ub){
						int[] fact = {i-1};
						satSolver.addClause(fact);
						ub--;
					}
				}
				if(usmode){
					for(int i:assumptions){
						int[] fact = {i};
						satSolver.addClause(fact);
						restVars--;
						offset++;
					}
					int a = assumptions[assumptions.length-1];
					for(int i=1; i<=margin+1; i++){
						int[] fact = {a-i};
						satSolver.addClause(fact);
						restVars--;
						offset++;
					}
				}
				System.out.println("c REFINE UPPERBOUND");
			} else { // UNSAT
				solveTime = getTime() - solveTime;
				System.out.println("c NOT FOUND: "+ objVar.getName() + " <= " + domValOfObjVar.get(border));
				System.out.println("c SOLVING: " + solveTime + " s");
				if(usmode){
					int[] cause = satSolver.getFailed(assumptions);
					int count = 0;
					for(int i: cause)
						if(i!=0)
							count++;
					if(count==0) break;
					int zeroVars = restVars-count+1;
					int start = vList.size()-restVars;
					for(int i=start; i<start+zeroVars; i++){	
						int[] fact = {vList.get(i).getCode()};
						satSolver.addClause(fact);
					}
					restVars -= zeroVars;
					int[] fact = {-(lbCode+lb+nonAssumedLits)};
					satSolver.addClause(fact);
					System.out.println("c REFINE LOWERBOUND");
					lb += zeroVars;
				} else
					break;
			}
		}
		
		if(isSAT){
			optSolveTime = getTime() - optSolveTime;
			System.out.println("c OBJECTIVE " + objVar.getName() + " " + objVar.getValue());
			System.out.println("s OPTIMUM FOUND");
			System.out.println("c Number of Solves : " + numberOfSolve);
			System.out.println("c Total Solving Time : " + optSolveTime + " s");
		} else {
			optSolveTime = getTime() - optSolveTime;
			System.out.println("s UNSATISFIABLE");
			System.out.println("c Number of Solves : " + numberOfSolve);
			System.out.println("c Total Solving Time : " + optSolveTime + " s");
		}
	}

	/**
	 * add channeling constraint to SAT solver.
	 * this method is called when selected "us2" strategy.
	 * @param objVar
	 *  objective variable for optimizing
	 * @throws SugarException
	 * @throws IOException
	 */
	private void addChannelingConstraint_forv2(IntegerVariable objVar) throws SugarException, IOException{
		int size = objVar.getDomain().size()-1;
		int lb = objVar.getDomain().getLowerBound();
		IntegerDomain bin = IntegerDomain.create(0, 1);
		for(int i=lb; i<lb+size; i++){
			IntegerVariable v = new IntegerVariable(bin);
			csp.add(v);
			vList.add(v);
		}
		ChannelingConstraintEncoder.encodeCC(new ArrayList<IntegerVariable>(vList), objVar);
		for(IntegerVariable v : ChannelingConstraintEncoder.newVars)
			csp.add(v);
		encodeDelta(ChannelingConstraintEncoder.expressions);
		ChannelingConstraintEncoder.clear();
	}

	/**
	 * add channeling constraint to SAT solver.
	 * this method is called when selected "us3" strategy.
	 * @param objVar
	 *  objective variable for optimizing
	 * @throws SugarException
	 * @throws IOException
	 */
	private void addChannelingConstraint_forv3(IntegerVariable objVar) throws SugarException, IOException{
		int size = objVar.getDomain().size()-1;
		int[] literals = new int[2*size];
		int lb = objVar.getDomain().getLowerBound();
		IntegerDomain bin = IntegerDomain.create(0, 1);
		for(int i=lb; i<lb+size; i++){
			IntegerVariable v = new IntegerVariable(bin);
			csp.add(v);
			vList.add(v);
		}
		Logger.info("Encoding additional constraint to SAT...");
		long start, stop;
		start = System.nanoTime();
		encodeDelta();
		for(int i=0; i<size; i++){
			literals[i] = (vList.get(i).getCode());
		}
		for(int i=size; i<2*size; i++){
			literals[i] = (objVar.getCode()+i-size);
		}
		if(satSolver instanceof SatSolverForNative){
			satSolver.newVar(size);
		}
		satSolver.addAtLeast(literals, size);
		//satSolver.addExactly(literals, size);
		stop = System.nanoTime();
		variablesCount = satSolver.getVars();
		clausesCount = satSolver.getClauses();
		Logger.info("SAT : " + summary());
		Logger.info("Additional Encoding Time : " + (stop-start) / 1000000000.0 + " sec");
	}
	
	
	@Override
	public void solveProblem() throws SugarException, IOException {
		if(strategy.equals("bin") || strategy.equals("mix"))
			optimize();
		else if(strategy.equals("us0") || strategy.equals("v0"))
			optimize_us_v0();
		else if(strategy.equals("us1") || strategy.equals("v1"))
			optimize_us_v1();
		else if(strategy.contains("us2") || strategy.contains("v2")){
			setAssumptionMode();
			optimize_us_v2();
		}
		else if(strategy.contains("us3") || strategy.contains("v3")){
			setAssumptionMode();
			optimize_us_v3();
		}
		else
			throw new SugarException("c Invalid strategy: " + strategy);
	}

}
