package iSugar;

import iSATLibrary.ISatSolver;
import iSATLibrary.SatSolverForNative;
import iSATLibrary.SatSolverForSat4j;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.BitSet;
import java.util.List;
import java.util.zip.GZIPInputStream;

import jp.kobe_u.sugar.Logger;
import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.converter.Converter;
import jp.kobe_u.sugar.converter.Simplifier;
import jp.kobe_u.sugar.csp.BooleanVariable;
import jp.kobe_u.sugar.csp.CSP;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.encoder.Encoder;
import jp.kobe_u.sugar.encoder.Problem;
import jp.kobe_u.sugar.expression.Expression;
import jp.kobe_u.sugar.expression.Parser;
import static iSugar.Solver.Status.*;

public abstract class Solver extends Problem {

	public class Status{
		private Status(){}
		public static final int SATISFIABLE = 10;
		public static final int UNSATISFIABLE = 20;
		public static final int INDETERMINATE = 0;
	}
	
	protected static CSP csp;
	protected static MaxCSP maxcsp;
	protected ISatSolver satSolver;
	protected Encoder encoder;
	protected int verbosity = 1;
	protected String solverName;
		
	protected static boolean isMaxCSP = false;
	protected static boolean incremental = true;
	protected static boolean enumeration = false;
	protected static boolean hcp = false;
	public static String strategy = "bin";
	protected static String problemFile;
	
	public Solver(String solverName){
		this.solverName = solverName;
		this.satSolver = solverName.equals("sat4j") ? new SatSolverForSat4j() : new SatSolverForNative(solverName);
	}
	
	public void addNormalizedClause(int[] lits) {
		satSolver.addNormalizedClause(lits);
	}

	public void addClauses(int[][] clauses) {
		satSolver.addClauses(clauses);
	}
	
	public void done(){
		satSolver.done();
	}
	
	 public void add_clause(int[] clause){
		 satSolver.addClause(clause);
	 }
	 
	 public void addAssumptions(int[] literals){
		 satSolver.addAssumptions(literals);
	 }
	
	 public int solveSAT(){
		 return satSolver.solve();
	 }

	 public int solveSAT(int[] assumption){
		 return satSolver.solve(assumption);
	 }	
	 
	 public void setTerminate(int state){
		 satSolver.setTerminate(state);
	 }

	 public void addBlockingClause(int[] literals){
		 satSolver.addBlockingClause(literals);
	 }
	
	public void addBlockingClause(){
		satSolver.addBlockingClause();
	}

	public void addPB(int[] weights, int[] literals, int k, boolean isgeq){
		variablesCount = satSolver.addPB(weights, literals, k, isgeq);
	}

	public void addPB(int[] weights, int[] literals, int k, boolean isgeq, int assumption){
		variablesCount = satSolver.addPBWithAssumption(weights, literals, k, isgeq, assumption);
	}
	
	public void addAtMost(int[] literals, int k){
		variablesCount = satSolver.addAtMost(literals, k);
	}
	
	public void addAtMost(int[] literals, int k, int assumption){
		variablesCount = satSolver.addAtMostWithAssumption(literals, k, assumption);
	}

	public void addAtLeast(int[] literals, int k){
		variablesCount = satSolver.addAtLeast(literals, k);
	}
	
	public void addAtLeast(int[] literals, int k, int assumption){
		variablesCount = satSolver.addAtLeastWithAssumption(literals, k, assumption);
	}
	
	public void addExactly(int[] literals, int k){
		variablesCount = satSolver.addExactly(literals, k);
	}
	
	public void addExactly(int[] literals, int k, int assumption){
		variablesCount = satSolver.addExactlyWithAssumption(literals, k, assumption);
	}

	public void setFrozen(int literal){
		satSolver.setFrozen(literal);
	}
	
	public boolean isEliminated(int literal){
		return satSolver.isEliminated(literal);
	}
	
	public BitSet getModel() {
		return satSolver.getModel();
	}
	
	public int[] getIntModel(){
		return satSolver.getIntModel();
	}
	
	public int[] getFailed(int[] literals){
		return satSolver.getFailed(literals);
	}
	
	public void outputModel(){
		satSolver.outputModel();
	}

	public double getTime() {
		return satSolver.getTime();
	}

	public String getSolverName() {
		 return satSolver.getSolverName();
	}

	public String getSolverSignature(){
		return satSolver.getSignature();
	}
	
	public void finalize() throws Throwable{
	    try {
	        super.finalize();
	      } finally {
	        satSolver.release();
	      }
	}	
	
    public String summary() {
        return variablesCount + " SAT variables, " + clausesCount + " SAT clauses";
    }
	
	public static List<Expression> parse(String cspFileName) throws SugarException, IOException {
        Logger.fine("Parsing " + cspFileName);
		InputStream in;
		if (cspFileName.endsWith(".gz")) {
			in = new GZIPInputStream(new FileInputStream(cspFileName));
		} else {
			in = new FileInputStream(cspFileName);
		}
		BufferedReader reader = new BufferedReader(new InputStreamReader(in, "UTF-8"));
		Parser parser = new Parser(reader);
		List<Expression> expressions = parser.parse();
		if(isMaxCSP)
			expressions = MaxCSP.relaxConstraint(expressions);
		reader.close();
		reader = null;
		in.close();
		in = null;
		parser = null;
		Runtime.getRuntime().gc();
        Logger.info("parsed " + expressions.size() + " expressions");
		if(isMaxCSP)
			System.out.println("c MAX CSP: " + MaxCSP.numOfSoftConstraints + " constraints");
		return expressions;
	}	
	
	public static void loadCSP() throws SugarException, IOException {
		List<Expression> expressions = parse(problemFile);
        Logger.fine("Converting to clausal form CSP");
		csp = isMaxCSP ? new MaxCSP() : new CSP();
		Converter converter = new Converter(csp);
		converter.convert(expressions);
		converter = null;
		expressions = null;
		Expression.clear();
		Runtime.getRuntime().gc();
        Logger.fine("CSP : " + csp.summary());
        Logger.fine("Propagation in CSP");
		csp.propagate();
        Logger.fine("CSP : " + csp.summary());
		if (csp.isUnsatisfiable()) {
            Logger.info("CSP is unsatisfiable after propagation");
            Logger.println("s UNSATISFIABLE");
			return;
		}
        Logger.fine("Simplifing CSP clauses by introducing new Boolean variables");
		Simplifier simplifier = new Simplifier(csp);
		simplifier.simplify();
		Runtime.getRuntime().gc();
        Logger.info("CSP : " + csp.summary());
	}

	public void encode() throws SugarException {
		long start = 0, stop = 0;
		Logger.info("Encoding CSP to SAT...");
		start = System.nanoTime();
		if(solverName.equals("sat4j")){
			encoder = new Encoder4j(csp);
			encoder.encode(this);
			if (strategy.contains("us2")){
				int n = csp.getObjectiveVariables().get(0).getDomain().size()-1;
				SatSolverForSat4j.buffer = ChannelingConstraintEncoder.calcVars(n);
			}
			else if (strategy.contains("us3")){
				int n = csp.getObjectiveVariables().get(0).getDomain().size()-1;
				SatSolverForSat4j.buffer = n;
			}
			satSolver.setVars(variablesCount);
			encoder.encode();
		} else {
			encoder = new Encoder(csp);
			encoder.encode(this);
		}
		stop = System.nanoTime();
		csp.commit();
		if(isMaxCSP)
			maxcsp = (MaxCSP) csp;
        Logger.info("SAT : " + summary());
		Logger.info("Encoding Time : " + (stop-start) / 1000000000.0 + " sec");
	 }
	
	public void encodeDelta() throws IOException, SugarException{
		encoder.encodeDelta();
	}
	
	public void encodeDelta(List<Expression> expressions) throws IOException, SugarException{
		long start = 0, stop = 0;
		Converter converter = new Converter(csp);
		converter.convert(expressions);
		converter = null;
		expressions = null;
		Expression.clear();
		Runtime.getRuntime().gc();
		Logger.info("Encoding additional constraint to SAT...");
		start = System.nanoTime();
		encodeDelta();
		stop = System.nanoTime();
        Logger.info("SAT : " + summary());
		Logger.info("Additional Encoding Time : " + (stop-start) / 1000000000.0 + " sec");
	}
		
	public int solveCSP() throws SugarException{
		return solveCSP(null);
	}
	
	public int solveCSP(int assumption[]) throws SugarException {
		if (csp.isUnsatisfiable()) {
			return UNSATISFIABLE;
		}
		int res = solveSAT(assumption);
		if (res==SATISFIABLE) {
			BitSet model = getModel();
			for (IntegerVariable v : csp.getIntegerVariables()) {
				v.decode(model);
			}
			for (BooleanVariable v : csp.getBooleanVariables()) {
				v.decode(model);
			}
		}
		return res;
	}
	
	/**
	 * output CSP solution.
	 */
	public void outputSolution(){		
		if(isMaxCSP){
			for (IntegerVariable vv : csp.getIntegerVariables()){
				String name = vv.getName();
				if(!vv.isAux() && !name.startsWith(MaxCSP.RELAXATION) && !name.startsWith(MaxCSP.AUXILIARY))
					System.out.println("a " + name + "\t" + vv.getValue());
			}
			for (BooleanVariable vv : csp.getBooleanVariables()){
				String name = vv.getName();
				if(!vv.isAux() && !name.startsWith(MaxCSP.RELAXATION) && !name.startsWith(MaxCSP.AUXILIARY))
					System.out.println("a " + name + "\t" + vv.getValue());
			}
			System.out.println("a ");			
		} else { 
			for (IntegerVariable vv : csp.getIntegerVariables())
				if(!vv.isAux()) System.out.println("a " + vv.getName() + "\t" + vv.getValue());
			for (BooleanVariable vv : csp.getBooleanVariables())
				if(!vv.isAux()) System.out.println("a " + vv.getName() + "\t" + vv.getValue());
			System.out.println("a ");
		}
	}
			
	public abstract void solveProblem() throws SugarException, IOException;
			
}
